
import unittest
import lifelib.pythlib.crt as crt

class TestCRT(unittest.TestCase):

    def test_mulinv(self):

        self.assertEqual(crt.mul_inv(105, 128), 89)
        self.assertEqual(crt.mul_inv(89, 128), 105)

    def test_chinese_remainder(self):

        moduli = [13, 14, 15]
        residues = [1729 % n for n in moduli]
        self.assertEqual(crt.chinese_remainder(moduli, residues), 1729)

if __name__ == '__main__':
    unittest.main()




from ._iwriter import iwriter

family = 2
mantissa = {0, 1}

def number_of_states(rulestring):

    return int(rulestring[1:rulestring.index('b')])

def bitplanes(rulestring):

    nstates = number_of_states(rulestring)
    return (2 if (nstates == 3) else len(bin(nstates - 3)))

def create_rule(rulestring):

    nstates = number_of_states(rulestring)

    if (nstates < 3):
        raise ValueError("Number of states must be at least 3")

    logstring = rulestring[rulestring.index('b'):]
    for iset in [['sse2'], ['sse2', 'avx'], ['sse2', 'avx', 'avx2']]:
        with open('ll_%s_%s.asm' % (iset[-1], logstring), 'w') as f:
            ix = genewriter(f, iset)
            ix.genlogic(logstring)

    with open('iterators_%s.h' % rulestring, 'w') as f:
        f.write('#pragma once\n')
        f.write('#include <stdint.h>\n')
        f.write('#include "../lifeconsts.h"\n')
        f.write('#include "../lifeperm.h"\n')
        f.write('#include "../eors.h"\n')
        f.write('namespace %s {\n\n' % rulestring.replace('-', '_'))

        for iset in [['sse2'], ['sse2', 'avx'], ['sse2', 'avx', 'avx2']]:
            iw = genewriter(f, iset)
            iw.write_function(rulestring, 20, 16)
            iw.write_iterator()

        gwrite_leaf_iterator(f, nstates)
        f.write('}\n')

class genewriter(iwriter):

    def write_function(self, rulestring, rowcount, dwidth):

        name = 'iterate_%s_%d_%d' % (self.besti, rowcount, dwidth)
        params = 'uint32_t * __restrict__ d, uint32_t * __restrict__ e, uint32_t * __restrict__ h, uint32_t * __restrict__ j'
        self.f.write('    void %s(%s) {\n' % (name, params))
        logstring = rulestring[rulestring.index('b'):]
        self.assemble(logstring, 0, rowcount, dwidth)
        self.f.write('            for (int i = 1; i < %d; i++) {\n' % (rowcount - 1))
        self.f.write('                e[i-1] &= (~h[i]);\n')
        self.f.write('                j[i] = d[i] & (~e[i-1]);\n')
        self.f.write('                d[i] = e[i-1];\n')
        self.f.write('            }\n')
        self.f.write('        return;\n')
        self.f.write('    }\n\n')

    def write_iterator(self):

        name = 'iterate_var_%s' % self.besti
        self.f.write('    void %s(uint32_t * __restrict__ d, uint32_t * __restrict__ h) {\n' % name)
        self.f.write('        uint32_t e[32];\n')
        self.f.write('        uint32_t j[32];\n')
        self.f.write('        iterate_%s_20_16(d+6, e+6, h+6, j+6);\n' % self.besti)
        self.f.write('        for (int i = 8; i < 24; i++) { h[i] = j[i]; }\n')
        self.f.write('        return;\n')
        self.f.write('    }\n\n')

def gwli_bsi(f, bsi, msi):

    f.write('            apg::z64_to_r32_%s(inleaves, d);\n' % bsi)
    f.write('            apg::z64_to_r32_%s(hleaves, h);\n' % bsi)
    f.write('            iterate_var_%s(d, h);\n' % bsi)
    f.write('            apg::r32_centre_to_z64_%s(d, outleaf);\n' % msi)
    f.write('            apg::r32_centre_to_z64_%s(h, outleaf2);\n' % msi)

def gwrite_leaf_iterator(f, nstates):

    name = 'iterate_var_leaf'
    params = 'uint64_t * inleaves, uint64_t * hleaves, uint64_t * outleaf'
    f.write('    bool %s(%s) {\n' % (name, params))
    f.write('        uint64_t outleaf2[4];')
    f.write('        int bis = apg::best_instruction_set();\n')
    f.write('        uint32_t d[32];\n')
    f.write('        uint32_t h[32];\n')
    f.write('        if (bis >= 10) {\n')
    gwli_bsi(f, 'avx2', 'avx2')
    f.write('        } else if (bis >= 9) {\n')
    gwli_bsi(f, 'avx', 'avx')
    f.write('        } else if (bis >= 7) {\n')
    gwli_bsi(f, 'sse2', 'sse4')
    f.write('        } else {\n')
    gwli_bsi(f, 'sse2', 'ssse3')
    f.write('        }\n')

    postprocess(f, nstates)
    f.write('        return false;\n')
    f.write('    }\n\n')

def postprocess(f, nstates):

    br = '' if (nstates == 3) else (bin(nstates - 3)[2:])[::-1]

    # We run 256 parallel binary counters in outleaf:
    f.write('        for (int i = 0; i < 4; i++) {\n')
    f.write('            uint64_t carry = outleaf[4+i];\n')
    for i, c in enumerate(br):
        f.write('            outleaf[%d+i] ^= carry;\n' % (4 * i + 8))
        f.write('            carry &= outleaf[%d+i];\n' % (4 * i + 8))
        if (c == '1'):
            f.write('            outleaf[%d+i] |= outleaf2[i];\n' % (4 * i + 8))
    f.write('            outleaf[4+i] ^= carry;\n')
    for i, c in enumerate(br):
        f.write('            outleaf[%d+i] ^= carry;\n' % (4 * i + 8))
    f.write('            outleaf[4+i] |= outleaf2[i];\n')
    f.write('        }\n')



'''
In this file, we demonstrate the usability of genera by implementing the
Generations variant of Larger than Life, reusing much of the existing code
from both Generations and Larger than Life.
'''

from .generations import family, mantissa, postprocess
from .ltl import get_ltl_params

def number_of_states(rulestring):

    return int(rulestring[1:rulestring.index('r')])

def bitplanes(rulestring):

    nstates = number_of_states(rulestring)
    return (2 if (nstates == 3) else len(bin(nstates - 3)))

def create_rule(rulestring):

    nstates = number_of_states(rulestring)

    if (nstates < 3):
        raise ValueError("Number of states must be at least 3")

    logstring = rulestring[rulestring.index('r'):]
    gparams = get_ltl_params(logstring)

    with open('iterators_%s.h' % rulestring, 'w') as f:
        f.write('#pragma once\n')
        f.write('#include <stdint.h>\n')
        f.write('#include "../ltl.h"\n')
        f.write('namespace %s {\n\n' % rulestring.replace('-', '_'))

        f.write('    bool iterate_var_leaf(uint64_t *inleaves, uint64_t *hleaves, uint64_t *outleaf) {\n\n')
        f.write('        apg::ltl_kernel(inleaves, outleaf')
        for g in gparams:
            f.write(', %d' % g)
        f.write(');\n')
        f.write('        outleaf[0] &= (~hleaves[3]); outleaf[1] &= (~hleaves[6]); outleaf[2] &= (~hleaves[9]); outleaf[3] &= (~hleaves[12]);\n')
        f.write('        uint64_t outleaf2[4] = {inleaves[3] & (~outleaf[0]), inleaves[6] & (~outleaf[1]), inleaves[9] & (~outleaf[2]), inleaves[12] & (~outleaf[3])};\n')
        postprocess(f, nstates)
        f.write('        return false;\n')
        f.write('    }\n')
        f.write('}\n')



from lifelib.cgold import describe
from lifelib import lifelib_dir
import os
import unittest

class TestCGoLd(unittest.TestCase):

    def test_genesis_block(self):

        blocks = list(describe.main(os.path.join(lifelib_dir, 'cgold', 'block0.cbc')))
        self.assertEqual(len(blocks), 1)
        block = blocks[0]
        seeds = [x.split(':')[-1].strip() for x in block.splitlines() if x.startswith('#C seed')]
        self.assertEqual(len(seeds), 1)
        seed = seeds[0]
        self.assertEqual(seed, 'OlCtB56rP1OdOQisR9mqQ1isR5CrQlnrwc6qF1n3a006SwRs442350544')

if __name__ == '__main__':
    unittest.main()



import os
from .lowlevel import *
from numbers import Integral
from .crt import chinese_remainder, large_primes
from .samples import download_synthesis as dsynth
from .samples import catagolue_url
from .registry import register_pattern_callback
from .idgun import determine_if_gun

def randfill(item, shape):
    '''
    Create a numpy array of a specified shape with specified probabilities
    of being each item.
    '''

    import numpy as np
    if isinstance(item, float):
        item = {0: 1.0 - item, 1: item}

    p = np.array(list(item.values()), dtype=float)
    p = p / np.sum(p)
    v = np.array(list(item.keys()), dtype=np.uint64)

    return np.random.choice(v, p=p, size=shape)


def interp(exes):
    '''
    Create a matrix for use when sampling an image.
    '''

    import numpy as np
    rx = np.round(exes)
    rmin = np.min(rx).astype(np.int64)
    rmax = np.max(rx).astype(np.int64)
    p = np.arange(rmin, rmax + 1)
    a = np.minimum(exes[1:], exes[:-1])
    b = np.maximum(exes[1:], exes[:-1])
    q = p.reshape((-1, 1))
    im = np.maximum(1.0 - np.maximum(a - q + 0.5, 0.0) - np.maximum(q - b + 0.5, 0.0), 0.0)
    im = im / im.sum(axis=0)

    # now expand up to match the dimension of exes:
    z = np.zeros_like(im[:, :1])
    im = 0.5 * (np.concatenate([im, z], axis=1) + np.concatenate([z, im], axis=1))

    return p, im


def colourise_image(x, hv, saturation=1.0):

    # normalise hue vector:
    import numpy as np
    hv = np.array(hv)
    hv = hv - hv * (1.0 / len(hv))
    denominator = np.max(np.abs(hv))
    multiplier = (saturation / denominator) if (denominator > 0) else 0.0
    hv *= multiplier

    x = 2.0 * x.reshape(x.shape + (1,)) - 1.0 # maps to range [-1, 1]
    slack = np.maximum(0.0, 1.0 - np.abs(x))
    x = x + hv * slack
    x = np.round((x + 1.0) * 127.5) # maps to [0, 255]
    return x.astype(np.uint8)


def blur_animation(x, steps):

    n = len(x)
    for _ in range(steps):
        x = [0.5 * x[i] + 0.5 * x[(i + 1) % n] for i in range(n)]
    return x


class Pattern(object):

    def __init__(self, session, ptr, owner):
        '''
        Do not call this method directly; instead, use lt.pattern(rle_or_apgcode, rulestring)
        '''

        self.ptr = ptr
        self.lifelib = session.lifelib
        self.session = session
        self.owner = owner
        self.comments = []

        session.pattern_ptrs.add(ptr)

    def __del__(self):

        if self.ptr in self.session.pattern_ptrs:
            self.lifelib('DeletePattern', self.ptr)
            self.session.pattern_ptrs.remove(self.ptr)

    def __repr__(self):

        logdiam = self.lifelib('GetDiameterOfPattern', self.ptr)
        beszel_index = self.lifelib('GetBeszelIndex', self.ptr)
        ulqoma_index = self.lifelib('GetUlqomaIndex', self.ptr)
        rule = self.getrule()

        return ("<Pattern(logdiam=%d, beszel_index=%d, ulqoma_index=%d, rule=%s) owned by %r>" % (logdiam,
            beszel_index, ulqoma_index, rule, self.owner))

    def digest(self):

        return self.lifelib('GetPatternDigest', self.ptr)

    def octodigest(self):

        return self.lifelib('GetPatternOctodigest', self.ptr)

        '''
        od = [self.digest()]
        for o in ["rot270", "rot180", "rot90", "flip_x", "flip_y", "swap_xy", "swap_xy_flip"]:
            od.append(self(o).digest())
        od = sorted(set(od))
        return sum([(2*i+1)*j for (i, j) in enumerate(od)]) & (2**64 - 1)
        '''

    def advance(self, numgens, exponent=0):

        self.lifelib.ensure_limit(exponent, numgens, 'Number of generations')

        temp = self

        while (numgens != 0):

            t = tzcount(numgens)
            if (t > 0):
                numgens = numgens >> t
                exponent = exponent + t

            max_num = 2 ** 30

            x = int(numgens if (abs(numgens) < max_num) else (numgens % max_num))

            newptr = self.lifelib('AdvancePattern', temp.ptr, x, exponent)
            temp = Pattern(self.session, newptr, self.owner)
            numgens = numgens - x

        return temp

    def _solid(self, state, exponent=0):

        if (state < 0) or (state >= (2**64)):
            raise ValueError("States must be integers in the interval [0, 2^64 - 1].")

        temp = self.owner.pattern("", self.getrule())

        while (state != 0):

            t = tzcount(state)
            if (t > 0):
                state = state >> t
                exponent = exponent + t

            max_num = 2 ** 30

            x = int(state % max_num)

            newptr = self.lifelib('GetSolidForPattern', self.ptr, x, exponent)
            temp += Pattern(self.session, newptr, self.owner)
            state = state - x

        return temp

    def _semisolid(self, flags):

        newptr = self.lifelib('GetSemisolidForPattern', self.ptr, flags)
        return Pattern(self.session, newptr, self.owner)

    def _boolean_mutable(self, other, op):

        self.lifelib('BooleanPatternMutable', self.ptr, other.ptr, op)
        return self

    def _bitshift(self, shift):

        newptr = self.lifelib('BitshiftPattern', self.ptr, shift)
        return Pattern(self.session, newptr, self.owner)

    def _boolean_immutable(self, other, op):

        newptr = self.lifelib('BooleanPatternImmutable', self.ptr, other.ptr, op)
        return Pattern(self.session, newptr, self.owner)

    def __iand__(self, other):
        return self._boolean_mutable(other, 0)

    def __ior__(self, other):
        return self._boolean_mutable(other, 1)

    def __ixor__(self, other):
        return self._boolean_mutable(other, 2)

    def __isub__(self, other):
        return self._boolean_mutable(other, 3)

    def __iadd__(self, other):
        return self._boolean_mutable(other, 1)

    def __imul__(self, other):
        return self._boolean_mutable(other, 4)

    def __imatmul__(self, other):
        return self._boolean_mutable(other, 7)

    def __and__(self, other):
        return self._boolean_immutable(other, 0)

    def __or__(self, other):
        return self._boolean_immutable(other, 1)

    def __xor__(self, other):
        return self._boolean_immutable(other, 2)

    def __sub__(self, other):
        return self._boolean_immutable(other, 3)

    def __add__(self, other):
        return self._boolean_immutable(other, 1)

    def __mul__(self, other):
        return self._boolean_immutable(other, 4)

    def __matmul__(self, other):
        return self._boolean_immutable(other, 7)

    def __lshift__(self, other):
        return self._bitshift(other)

    def __rshift__(self, other):
        return self._bitshift(-other)

    def convolve(self, other, exclusive=False):
        return self._boolean_immutable(other, (6 if exclusive else 5))

    def disjunct(self, other, exclusive=False):
        return self._boolean_immutable(other, (2 if exclusive else 1))

    def __pow__(self, other):
        if (other <= 0):
            raise ValueError("Can only raise patterns to positive integer powers")
        elif (other == 1):
            return self
        else:
            return (self * (self ** (other - 1)))

    # These implement the 'subset' partial order defined on sets:

    def __eq__(self, other):
        return bool(self.lifelib('PatternEquality', self.ptr, other.ptr))

    def __ne__(self, other):
        return not (self == other)

    def __le__(self, other):
        return ((self & other) == self)

    def __ge__(self, other):
        return ((self & other) == other)

    def __lt__(self, other):
        return (self != other) and (self <= other)

    def __gt__(self, other):
        return (self != other) and (self >= other)

    def nonempty(self):
        return bool(self.lifelib('PatternNonempty', self.ptr))

    def empty(self):
        return not self.nonempty()

    __nonzero__ = nonempty
    __bool__ = nonempty

    def stream(self, data_or_filename):

        if isinstance(data_or_filename, list):
            data_or_filename = str(data_or_filename)

        newptr = self.lifelib('MakeSpaceshipStream', self.ptr, data_or_filename)
        return Pattern(self.session, newptr, self.owner)

    def transform(self, tfm):

        transforms = ["flip", "rot180", "identity", "transpose", "flip_x", "flip_y",
                        "rot90", "rot270", "swap_xy", "swap_xy_flip", "rcw", "rccw"]

        if tfm not in transforms:
            raise ValueError("Transformation must be one of %s" % str(transforms))

        newptr = self.lifelib('TransformPattern', self.ptr, tfm)
        return Pattern(self.session, newptr, self.owner)

    def __copy__(self):

        newptr = self.lifelib('CopyPattern', self.ptr)
        return Pattern(self.session, newptr, self.owner)

    def from_rle(self, pattern_or_rle):
        if isinstance(pattern_or_rle, str):
            pattern_or_rle = self.owner.pattern(pattern_or_rle, self.getrule())
        return pattern_or_rle


    def shift(self, x, y, exponent=0):

        if not isinstance(x, Integral):
            raise TypeError("shift arguments must be integers")

        if not isinstance(y, Integral):
            raise TypeError("shift arguments must be integers")

        self.lifelib.ensure_limit(exponent, x, 'Horizontal shift')
        self.lifelib.ensure_limit(exponent, y, 'Vertical shift')

        temp = self

        while ((x | y) != 0):

            t = tzcount(x | y)
            if (t > 0):
                x = x >> t
                y = y >> t
                exponent = exponent + t

            max_num = 2 ** 30
            dx = int(x if (abs(x) < max_num) else (x % max_num))
            dy = int(y if (abs(y) < max_num) else (y % max_num))
            newptr = self.lifelib('ShiftPattern', temp.ptr, dx, dy, exponent)
            temp = Pattern(self.session, newptr, self.owner)
            x -= dx
            y -= dy

        return temp

    def centre(self):

        if self.empty():
            return self

        bbox = self.getrect()
        return self.shift(-(bbox[0] + (bbox[2] // 2)), -(bbox[1] + (bbox[3] // 2)))

    def _getbound(self, direction, pixelsize=None, offset=None):

        if pixelsize is None:
            logdiam = self.lifelib('GetDiameterOfPattern', self.ptr)
            pixelsize = (logdiam // 30) * 30
            offset = -(1 << (logdiam - 1))

        b = int(self.lifelib('GetPatternBound', self.ptr, direction, pixelsize)) % (2 ** 30)

        if (pixelsize == 0):
            return b + offset

        b = b << pixelsize

        x = self(-b, 0) if (direction % 2 == 0) else self(0, -b)
        x = x[offset:offset + (1 << pixelsize),:] if (direction % 2 == 0) else x[:,offset:offset + (1 << pixelsize)]

        return b + x._getbound(direction, pixelsize - 30, offset)

    def component_containing(self, seed=None, halo='ooo$ooo$ooo!'):

        if seed is None:
            seed = self.onecell()

        if not isinstance(halo, Pattern):
            halo = self.from_rle(halo).centre()

        if isinstance(seed, tuple):
            seed = self.from_rle('o!')(seed[0], seed[1])

        newptr = self.lifelib('FindConnectedComponent', seed.ptr, self.ptr, halo.ptr)
        return Pattern(self.session, newptr, self.owner)

    def onecell(self):

        newptr = self.lifelib('GetOneCell', self.ptr)
        return Pattern(self.session, newptr, self.owner)

    def layers(self):

        return self.owner.deunify(self)

    def components(self, halo='ooo$ooo$ooo!'):

        ccs = []
        x = self

        while (x.nonempty()):

            cc = x.component_containing(halo=halo)
            x = x - cc
            ccs.append(cc)

        return ccs

    def match(self, live, dead=None, halo=None):

        # accept RLEs as well as patterns:
        live = self.from_rle(live)
        dead = self.from_rle(dead)

        if halo is not None:
            if not isinstance(halo, Pattern):
                halo = self.from_rle(halo).centre()
            dead = live.convolve(halo)

        if dead is None:
            newptr = self.lifelib('MatchLive', self.ptr, live.ptr)
        else:
            corona = dead - live
            newptr = self.lifelib('MatchLiveAndDead', self.ptr, live.ptr, corona.ptr)
        return Pattern(self.session, newptr, self.owner)

    def replace(self, live, replacement, dead=None, n_phases=0, stepsize=1, orientations=['identity'], **kwargs):

        # accept RLEs as well as patterns:
        live = self.from_rle(live)
        dead = self.from_rle(dead)
        replacement = self.from_rle(replacement)

        if n_phases > 0:
            temp = self
            for _ in range(n_phases):
                temp = temp[stepsize].replace(live, replacement, dead=dead, orientations=orientations, **kwargs)
            return temp

        if orientations == 'rotate4reflect':
            orientations = ["rot270", "rot180", "rot90", "identity",
                            "flip_x", "flip_y", "swap_xy", "swap_xy_flip"]

        replaced = self.from_rle("")
        matched  = self.from_rle("")

        for o in orientations:

            ro = replacement.transform(o)
            lo = live.transform(o)
            do = dead.transform(o) if (dead is not None) else dead

            m = self.match(lo, dead=do, **kwargs)
            matched  += m.convolve(lo)
            replaced += m.convolve(ro)

        return (self - matched) + replaced

    def __call__(self, *args):

        if (len(args) > 3):
            raise TypeError("Usage: pattern(8, -5) or pattern('rccw') or pattern('rot90', 7, 3)")

        temp = self.__copy__()

        if ((len(args) % 2) == 1):
            temp = temp.transform(args[0])

        if (len(args) >= 2):
            temp = temp.shift(args[-2], args[-1])

        return temp

    def pdetect_or_advance(self, maxexp=24):

        newptr = self.lifelib('FindPeriodOrAdvance', self.ptr, maxexp)

        if newptr[0]:
            return Pattern(self.session, newptr, self.owner)
        else:
            dt = self.lifelib('GetDTOfPattern', self.ptr)
            dx = self.lifelib('GetDXOfPattern', self.ptr)
            dy = self.lifelib('GetDYOfPattern', self.ptr)
            return {'period': dt, 'displacement': (dx, dy)}

    def getrule(self):

        return self.lifelib('GetRuleOfPattern', self.ptr, [2048])

    def phase_wechsler(self, fn='GetWechslerOfPattern'):

        x = self.lifelib(fn, self.ptr, [2048], 2048)
        if x.startswith('!'):
            buflen = int(x[1:])
            x = self.lifelib(fn, self.ptr, [buflen], buflen)
        return x

    def oscar(self, maxexp=16, eventual_oscillator=True, verbose=True, return_apgcode=False, allow_guns=True):

        logdiam = self.lifelib('GetDiameterOfPattern', self.ptr)

        if (logdiam > 60):
            raise ValueError("Pattern is too large for oscillation detection.")

        x = self
        for exponent in range(8, maxexp+2, 2):
            if (verbose) and (exponent > 8):
                print("Checking periodicity with exponent %d..." % exponent)
            y = x.pdetect_or_advance(exponent)
            if isinstance(y, dict):
                if (return_apgcode):
                    y['apgcode'] = x.phase_wechsler('GetApgcodeOfPattern')
                return y
            if (eventual_oscillator):
                x = y

        if allow_guns and self.getrule() == 'b3s23':
            rv = determine_if_gun(self, maxexp)
            if rv is not None:
                cg, multiple, cw, z, envelope = rv
                return {'period': multiple, 'displacement': tuple([-x for x in z.displacement]), 'apgcode': 'xg%d_%s' % (multiple, cw)}

        if (verbose):
            print("Oscillation not detected (try increasing maxexp?)")
        return {}

    def write_rle(self, filename, header=None, footer=None, comments=None, file_format='rle', save_comments=True):

        filename = os.path.abspath(filename)

        if header is None:
            header = ''

        if footer is None:
            footer = ''

        if save_comments:
            if comments is None:
                comments = self.comments
            if hasattr(comments, 'splitlines'):
                comments = comments.splitlines()
            comments = [(x if x.startswith('#C ') else ('#C ' + x)) for x in comments]
            header += ''.join([('%s\n' % x) for x in comments])

        llc = 'SavePatternRLE' if (file_format[-3:].lower() == 'rle') else 'SavePatternMC'
        self.lifelib(llc, self.ptr, filename, header, footer)

    def write_file(self, filename, file_format='deduce', compressed=False, tempfile='tempfile', **kwargs):

        if file_format == 'deduce':
            compressed = (filename[-3:] == '.gz')
            filebase = filename[:-3] if compressed else filename
            file_format = filebase[-3:]
        elif file_format[-3:] == '.gz':
            compressed = True
            file_format = file_format[:-3]

        if compressed:
            self.write_rle(tempfile, file_format=file_format, **kwargs)
            import gzip
            import shutil
            with open(tempfile, 'rb') as f_in:
                with gzip.open(filename, 'wb') as f_out:
                    shutil.copyfileobj(f_in, f_out)
            try:
                os.remove(tempfile)
            except OSError:
                pass
        else:
            self.write_rle(filename, file_format=file_format, **kwargs)

    # This seems to have better symmetry with Lifetree.load() so we
    # include it as an alias for Pattern.write_file():
    save = write_file


    def rle_string(self, filename=None, autoremove=True):

        return self.phase_wechsler('GetRleOfPattern')


    def viewer(self, filename=None, width=480, height=480, base64=True,
                lv_config='#C [[ THEME 6 GRID GRIDMAJOR 0 ]]', autoremove=True, edit=True):

        if filename is None:
            filename = self.session.newfloat('viewer') + '.html'

        header = '<html><head><meta name="LifeViewer" content="rle code"></head><body>'
        header += '<div class="rle"><div style="display:none;"><code id="code2">\n'
        footer = '#C [[ WIDTH %d HEIGHT %d ]]\n' % (width, height)
        footer += '</code></div>\n<canvas width="%d" height="%d" style="margin-left:1px;"></canvas></div>\n' % (width+16, height+16)
        footer += "<script type='text/javascript' src='%s/js/lv-plugin.js'></script>\n" % catagolue_url

        if edit:
            # Generate a unique identifier for the IFrame, so that the Jupyter
            # notebook knows which LifeViewer has been updated with Ctrl+S.
            from uuid import uuid4
            saveid = str(uuid4())

            # Track when the (invisible!) RLE element is changed, and signal
            # those changes in a message from the IFrame to the notebook. We
            # can catch these messages in the notebook itself.
            footer += '''<script>
var targetNode = document.getElementById('code2');
var config = { attributes: true, childList: true, subtree: true };
var callback = function(mutationsList, observer) {
    console.log("DOM mutated.");
    parent.postMessage({ rle: targetNode.innerHTML, uuid: "%s" }, "*");
};
var observer = new MutationObserver(callback);
observer.observe(targetNode, config);
</script>''' % saveid

            # Register this pattern in the registry:
            register_pattern_callback(saveid, self)

        footer += '</body></html>\n'

        self.write_rle(filename, header, lv_config + '\n' + footer)

        if base64:

            from base64 import b64encode

            with open(filename, 'rb') as f:
                b64html = regstring(b64encode(f.read()))
            source = 'data:text/html;base64,%s' % b64html

            if autoremove:
                try:
                    os.remove(filename)
                except OSError:
                    pass

        else:
            source = filename

        from IPython.display import IFrame

        return IFrame(source, width=width+32, height=height+32)

    def __getitem__(self, x):

        if isinstance(x, Integral):
            return self.advance(x)
        elif isinstance(x, tuple) and isinstance(x[0], Integral) and isinstance(x[1], Integral):
            left = x[0]
            top = x[1]
            relevant_cell = self[left:left+1,top:top+1].shift(-left, -top)
            return self.lifelib('GetOriginState', relevant_cell.ptr)
        elif isinstance(x, tuple):
            left = x[0].start
            right = x[0].stop
            top = x[1].start
            bottom = x[1].stop

            subrect = self
            bigenough = self

            onecell = self.from_rle("o!")

            if left is not None:
                bigenough = bigenough + onecell.shift(left, 0)
            if right is not None:
                bigenough = bigenough + onecell.shift(right-1, 0)
            if top is not None:
                bigenough = bigenough + onecell.shift(0, top)
            if bottom is not None:
                bigenough = bigenough + onecell.shift(0, bottom-1)

            if left is not None:
                subrect = subrect & bigenough._semisolid(10).shift(left, 0)
            if right is not None:
                subrect = subrect & bigenough._semisolid(5).shift(right, 0)
            if top is not None:
                subrect = subrect & bigenough._semisolid(12).shift(0, top)
            if bottom is not None:
                subrect = subrect & bigenough._semisolid(3).shift(0, bottom)

            return subrect
        elif hasattr(x, 'ctypes'):

            logdiam = self.lifelib('GetDiameterOfPattern', self.ptr)
            if (logdiam >= 64):
                # Ensure we can safely use int64 arithmetic:
                self64 = self[-(2**63):(2**63), -(2**63):(2**63)]
            else:
                self64 = self

            import numpy as np
            x = np.ascontiguousarray(x, dtype=np.int64)
            n = len(x)
            y = np.zeros((n,), dtype=np.uint64, order='C')
            y = self.lifelib('GetCells', self64.ptr, n, x, y)
            return y
        else:
            raise TypeError("operator[] accepts either a generation (int) or a pair of slices")

    def _subpops(self, n, ps):

        if (ps < 4):
            raise ValueError('log2 of pixel size must be at least 4.')

        logdiam = self.lifelib('GetDiameterOfPattern', self.ptr)
        il = n + ps
        if (logdiam > il):
            self64 = self[-(2**(il-1)):(2**(il-1)), -(2**(il-1)):(2**(il-1))]
        else:
            self64 = self

        import numpy as np
        y = np.zeros((2**n, 2**n), dtype=np.uint64, order='C')
        y = self.lifelib('GetSubpops', self64.ptr, n, ps, y)
        return y

    def coords(self):

        p = self.population
        import numpy as np
        y = np.zeros((p,2), dtype=np.int64, order='C')
        y = self.lifelib('GetCoords', self.ptr, y)
        return y

    def destream(self, other):

        import numpy as np
        oscr = self.oscar()
        disp = np.array(list(oscr['displacement']))
        p = oscr['period']

        offsets = []

        for i in range(p):
            c = other.match(self[i], halo="b3o$5o$5o$5o$b3o!").coords()
            offsets.append(np.dot(c, disp) * p + i * np.dot(disp, disp))

        offsets = np.concatenate(offsets)
        offsets.sort() # in-place
        offsets = np.diff(offsets)[::-1] // np.dot(disp, disp)
        return ([0] + list(offsets) + [0])

    def __setitem__(self, x, item):

        if hasattr(x, 'ctypes'):
            import numpy as np
            x = np.ascontiguousarray(x, dtype=np.int64)
            n = len(x)
            y = np.ascontiguousarray(item, dtype=np.uint64)

            if (len(x) != len(y)):
                raise ValueError('Incompatible shapes for setcells.')

            self.lifelib('SetCells', self.ptr, n, x, y)
            return

        x = tuple([(slice(y, y + 1) if isinstance(y, Integral) else y) for y in x])

        self -= self[x]

        if isinstance(item, float) or isinstance(item, dict):
            shape = (x[0].stop - x[0].start, x[1].stop - x[1].start)
            item = randfill(item, shape)

        if hasattr(item, 'ctypes'):
            import numpy as np
            xr = np.arange(x[0].start, x[0].stop)
            yr = np.arange(x[1].start, x[1].stop)
            coords = np.ascontiguousarray(np.array(list(map(np.ravel, np.meshgrid(xr, yr)))).T, dtype=np.int64)
            values = np.ascontiguousarray(np.ravel(item), dtype=np.uint64)
            if (len(coords) != len(values)):
                raise ValueError('Incompatible shapes for setcells.')

            self.lifelib('SetCells', self.ptr, len(coords), coords, values)
            return
        elif isinstance(item, Integral):
            if (item == 0):
                return
            onecell = self.from_rle("o!")
            bigenough = onecell(x[0].start, x[1].start) + onecell(x[0].stop - 1, x[1].stop - 1)
            item = bigenough._solid(item)

        self += item[x]

    def getrect(self):

        if self.empty():
            # An empty pattern does not possess a well-defined bounding box:
            return None

        logdiam = self.lifelib('GetDiameterOfPattern', self.ptr)

        if (logdiam > 60):
            # Slower bigint solution:
            bounds = [self._getbound(i) for i in range(4)]
            bbox = [bounds[0], bounds[1], bounds[2] + 1 - bounds[0], bounds[3] + 1 - bounds[1]]
        else:
            # Fast int64 solution:
            bbox = self.lifelib('GetPatternBox', self.ptr, [1, 4])

        return bbox

    @property
    def bounding_box(self):

        return self.getrect()

    @property
    def population(self):

        logdiam = self.lifelib('GetDiameterOfPattern', self.ptr)
        nprimes = (logdiam + 14) // 15
        relprimes = large_primes[:nprimes]
        populations = [self.lifelib('GetPopulationOfPattern', self.ptr, p) for p in relprimes]
        return chinese_remainder(relprimes, populations)

    @property
    def wechsler(self):

        return self.phase_wechsler()

    @property
    def apgcode(self):

        return self.oscar(return_apgcode=True, eventual_oscillator=False, verbose=False)['apgcode']

    def download_samples(self, **kwargs):
        '''
        Download a dictionary mapping a symmetry string to a Sequence of sample soups.
        '''

        return self.owner.download_samples(self.apgcode, self.getrule(), **kwargs)

    def download_synthesis(self, tempfile='tempfile.rle', **kwargs):
        '''
        Download a synthesis RLE of this pattern from Catagolue, or return None
        if no such synthesis exists.
        '''

        rle = dsynth(self.apgcode, self.getrule(), tempfile=tempfile, **kwargs)

        if rle is None:
            synth = None
        else:
            synth = self.owner.load(tempfile)

        try:
            os.remove(tempfile)
        except OSError:
            pass

        return synth

    @property
    def period(self):

        return self.oscar(eventual_oscillator=False, verbose=False)['period']

    @property
    def displacement(self):

        return self.oscar(eventual_oscillator=False, verbose=False)['displacement']

    def sample(self, exes, whys):

        import numpy as np

        x, mx = interp(exes)
        y, my = interp(whys)

        xx, yy = np.meshgrid(x, y)
        c = np.concatenate([xx.reshape((-1,1)), yy.reshape((-1,1))], axis=1)
        zz = self[c].reshape(xx.shape)

        return np.dot(my.T, np.dot(zz, mx))

    def smooth_step(self, frames, gens):

        currgen = 0
        y = self

        for i in range(frames):
            nextgen = (i * gens) // frames
            if nextgen != currgen:
                y = y[nextgen - currgen]
                currgen = nextgen

            yield y

    def voxels(self, frames, gens, displacement):

        import numpy as np
        gd = np.array(list(displacement)) * (1.0 / frames)
        v = [y.coords() - gd * i for i, y in enumerate(self.smooth_step(frames, gens))]
        return v

    def animate(self, frames, gens, displacement, params, width, height):
        '''
        frames: number of frames in the output animation
        gens: number of generations to simulate
        displacement: amount to move the camera throughout the course of the animation
        params: a triple (x_centre, y_centre, cells_per_pixel)
        width: image width (in pixels)
        height: image height (in pixels)
        '''

        import numpy as np
        v = []
        ix = params[2] * np.arange(width)
        iy = params[2] * np.arange(height)
        ix = ix + params[0] - 0.5 * ix[-1]
        iy = iy + params[1] - 0.5 * iy[-1]
        dx = displacement[0] * (1.0 / frames)
        dy = displacement[1] * (1.0 / frames)

        return [y.sample(ix + dx * i, iy + dy * i) for i, y in enumerate(self.smooth_step(frames, gens))]

    def centre_and_pixel_size(self, gens, displacement, width, height):
        '''
        Emits a triple (x_centre, y_centre, cells_per_pixel) suitable for
        making an animation that ensures that all live cells in all
        generations stay within the image boundary.
        '''

        import numpy as np
        v = np.concatenate(self.voxels(gens, gens, displacement), axis=0)

        minv = np.min(v, axis=0)
        maxv = np.max(v, axis=0)
        centre = 0.5 * (minv + maxv)

        dims = (4.0 + maxv - minv)
        pixelsize = max(dims[0] / width, dims[1] / height)

        return float(centre[0]), float(centre[1]), float(pixelsize)

    def make_gif_internal(self, p, disp, hue='#ff0000', width=350, height=350, border=5, bg=None):

        import numpy as np
        saturation = 1.0

        if isinstance(hue, str):

            hue = hue.lower().lstrip('#')
            lh = len(hue)
            if (lh % 3) != 0:
                raise ValueError("invalid hexadecimal colour code")
            lh = lh // 3
            hue = np.array([int(hue[lh*i:lh*(i+1)], 16) for i in range(3)])
            hue = hue / (16 ** lh - 1.0)
            saturation = np.max(hue) - np.min(hue)

        if bg is not None:
            params = bg.centre_and_pixel_size(1, (0, 0), width - 2 * border, height - 2 * border)
        else:
            params = self.centre_and_pixel_size(p, disp, width - 2 * border, height - 2 * border)

        frames_per_gen = max(2, min(240 // p, 8))
        frames = frames_per_gen * p

        anim = self.animate(frames, p, disp, params, width, height)
        anim2 = [colourise_image(x, hue, saturation) for x in blur_animation(anim, 2)]

        return anim2
        

    def make_gif(self, filename=None, fps=25, **kwargs):
        '''
        Intelligently create an animated GIF of an oscillator or spaceship.
        Spaceships will be oriented to point between north and north-west.
        '''

        y = self
        d = y.oscar(maxexp=16, return_apgcode=True, verbose=False)

        disp = list(d['displacement'])
        p = d['period']

        if abs(disp[0]) > abs(disp[1]):
            y = y.transform('swap_xy')
            disp = disp[::-1]

        if disp[0] > 0:
            y = y.transform('flip_x')
            disp[0] = -disp[0]

        if disp[1] > 0:
            y = y.transform('flip_y')
            disp[1] = -disp[1]

        bg = None

        if d['apgcode'].startswith('xg'):
            disp = (0, 0)
            bg = determine_if_gun(y, 16)[4]
            rect = bg.getrect()
            periods = 1 + (4 * (rect[2] + rect[3])) // p
            y = y[p * periods]

            if (rect[3] > rect[2]):
                y = y.transform('swap_xy')
                bg = bg.transform('swap_xy')

        gif = y.make_gif_internal(p, disp, bg=bg, **kwargs)

        if filename is None:
            return gif

        import imageio
        imageio.mimsave(filename, gif, fps=fps)

        # return the pattern that was giffed:
        return y



from .parsetable import GetNumberOfInputs
from .parsetree import normalise_neighbourhood


def MakeStaticTable(nodes):

    # Write rule tree in .data segment of the program:
    lines = ["const static uint32_t __ruletree[] __attribute__((aligned(64))) = {\n"]
    for (i, l) in enumerate(nodes):
        body = '    ' + (','.join([('%du' % x) for x in l]))
        body += ('};\n\n' if (i == len(nodes) - 1) else ',\n')
        lines.append(body)
    lines.append('\n')

    return lines


def NodesToCode(nhood, is16bit):

    numInputs = GetNumberOfInputs(nhood)
    numOutputs = len(nhood) - numInputs

    if (numOutputs != 1):
        raise RuntimeError("NodesToCode currently supports at most one output")

    nhood = normalise_neighbourhood(nhood)

    for (x, y) in nhood:
        if (x > 8) or (x < -8) or (y > 8) or (y < -8):
            raise RuntimeError("Neighbourhood must be a subset of [-8, 8] x [-8, 8]")

    # Byte offsets to beginnings of rows:
    in_offsets  = [(y+8)*32 + (x+8) for (x, y) in nhood[:numInputs]]
    out_offsets = [ y * 16  +   x   for (x, y) in nhood[numInputs:]]

    if is16bit:
        out_offsets += [t + 256 for t in out_offsets]

    # Hopefully simple enough to vectorise:
    lines = ['void iterate_var_row(uint8_t* __restrict__ inrow, uint8_t* __restrict__ outrow) {\n']
    lines.append('    uint32_t temprow[16] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};\n')
    for offset in in_offsets:
        if (is16bit):
            offset2 = ' + %d' % (offset + 1024)
            lines.append('    for (int i = 0; i < 16; i++) { temprow[i] += (((uint32_t) inrow[i%s]) << 8); }\n' % offset2)
        offset = '' if (offset == 0) else (' + %d' % offset)
        lines.append('    for (int i = 0; i < 16; i++) { temprow[i] += inrow[i%s]; }\n' % offset)
        lines.append('    for (int i = 0; i < 16; i++) { temprow[i] = __ruletree[temprow[i]]; }\n')
    for (i, offset) in enumerate(out_offsets):
        offset = '' if (offset == 0) else (' + %d' % offset)
        shift = '' if (i == 0) else (' >> %d' % (8*i))
        lines.append('    for (int i = 0; i < 16; i++) { outrow[i%s] = temprow[i]%s; }\n' % (offset, shift))
    lines.append('}\n')
    lines.append('\n')

    return lines



import os
import re

from collections import Sequence, defaultdict

catagolue_url = "https://catagolue.hatsya.com"

def get_symmetries(rulestring):

    validsyms = ["1x256", "2x128", "4x64", "8x32", "C1"] # asymmetric soups
    for sz in ["32x32", "64x64", "128x128", "256x256", "512x512", "1k", "2k", "4k", "8k", "Glider8_4_5", "Glider6_5_6"]:
        validsyms += ["Mateon1_" + sz + "_Test"]

    if rulestring[-1].lower() == 'h':
        validsyms += ["C2_4", "C2_1", "C3_1", "C6", "D2_x", "D2_xo", "D4_x4", "D4_x1", "D6_1", "D6_1o", "D12"]
    else:
        validsyms += ["C2_4", "C2_2", "C2_1", "C4_4", "C4_1",
                        "D2_+2", "D2_+1", "D2_x", "D4_+4", "D4_+2", "D4_+1", "D4_x4", "D4_x1", "D8_4", "D8_1", "D2_+1_gO1s0", "D2_+1_gO1s1", "D2_+1_gO1s2"]

    if (re.match('^b0?1?2?3?4?5?6?7?8?s0?1?2?3?4?5?6?7?8?$', rulestring)):
        validsyms += ["G1", "H2_+1", "H2_+2", "H4_+1", "H4_+2", "H4_+4"] # GPU symmetries

    return validsyms

def validate_symmetry(rulestring, symmetry, throw_error=True):

    if (len(symmetry) > 50):
        if throw_error:
            raise ValueError("%s exceeds maximum of 50 characters" % symmetry)
        return False
    elif (len(symmetry) < 2):
        if throw_error:
            raise ValueError("%s is less than minimum of 2 characters" % symmetry)
        return False

    if not (re.compile("^[a-zA-Z0-9_+]+$").match(symmetry)):
        if throw_error:
            raise ValueError("%s contains an invalid character (not in [a-zA-Z0-9_+])" % symmetry)
        return False

    if "stdin" in symmetry:
        return True

    validsyms = get_symmetries(rulestring)

    redsym = symmetry

    if 'D8_1' in validsyms:
        while ((len(redsym) > 0) and (redsym[0] == 'i')):
            redsym = redsym[1:]

    if throw_error and (redsym not in validsyms):
        raise ValueError("%s is not one of the supported symmetries: %s" % (symmetry, validsyms))

    return (redsym in validsyms)

class SampleSoupList(Sequence):

    def __init__(self, lt, rule, symmetry, samples):

        self.lt = lt
        self.rule = rule
        self.symmetry = symmetry
        self.samples = samples

    def __getitem__(self, i):
        return self.lt.hashsoup(self.rule, self.symmetry, self.samples[i])

    def __len__(self):
        return len(self.samples)

    def __repr__(self):
        return 'SampleSoupList(%r, %r, %r, %r)' % (self.lt, self.rule, self.symmetry, self.samples)

def download_synthesis(apgcode, rule, domain=catagolue_url, tempfile='tempfile'):

    if ('_' in rule) and ('_' not in apgcode):
        rule, apgcode = apgcode, rule

    try:
        from urllib import urlretrieve
    except ImportError:
        from urllib.request import urlretrieve

    url = domain + '/textsamples/' + apgcode + '/' + rule + '/synthesis'
    urlretrieve(url, tempfile)

    with open(tempfile, 'r') as f:
        synthesis = f.read()

    if 'x' not in synthesis:
        return None
    else:
        return synthesis

def download_samples(lt, apgcode, rule, domain=catagolue_url, tempfile='tempfile'):

    if ('_' in rule) and ('_' not in apgcode):
        rule, apgcode = apgcode, rule

    try:
        from urllib import urlretrieve
    except ImportError:
        from urllib.request import urlretrieve

    url = domain + '/textsamples/' + apgcode + '/' + rule
    urlretrieve(url, tempfile)

    samples = defaultdict(list)

    with open(tempfile, 'r') as f:
        for l in f:
            if '/' not in l:
                continue
            symmetry, seed = tuple(l.strip().split('/')[:2])
            if (len(symmetry) > 50) or (len(symmetry) == 0):
                continue
            if (symmetry[0] == 'G'):
                symmetry = 'C' + symmetry[1:]
            elif (symmetry[0] == 'H'):
                symmetry = 'D' + symmetry[1:]
            samples[symmetry].append(seed)

    samples = {k : SampleSoupList(lt, rule, k, v) for (k, v) in samples.items() if validate_symmetry(rule, k, False)}
    return samples



import unittest
import lifelib
import os

class TestGoL(unittest.TestCase):

    @classmethod
    def setUpClass(cls):

        lifelib.load_rules('b3s23', force_compile=True)

    def setUp(self):

        self.sess = lifelib.load_rules('b3s23')
        self.lt = self.sess.lifetree(memory=1000, n_layers=4)

    def test_bitshift(self):

        pd = self.lt.pattern('xp15_4r4z4r4')
        self.assertEqual((pd << 3).wechsler, '0_0_0_4r4z4r4')
        self.assertEqual((pd << 2).wechsler, '0_0_4r4z4r4')
        self.assertEqual((pd << 1).wechsler, '0_4r4z4r4')

        self.assertEqual(((pd << 2) >> 2), pd)

        block = self.lt.pattern('xs4_33')
        hwss = self.lt.pattern('xq4_27deee6')
        mwss = self.lt.pattern('xq4_27dee6')

        combined = self.lt.unify(block, pd, hwss, mwss)
        self.assertEqual(combined.wechsler, '33_4r4z4r4_27deee6_27dee6')
        self.assertEqual(combined.layers(), (block, pd, hwss, mwss))

    def test_ulimit(self):

        self.sess.lifelib.ensure_limit(10000, 1)
        self.sess.lifelib.ensure_limit(20000, 1)
        self.sess.lifelib.ensure_limit(10000, 1)

    def test_freywa_bug(self):

        a = self.lt.pattern("bo$2bo$3o5$5bobo$6b2o$6bo")
        b = self.lt.pattern("bo$2bo$3o")
        self.assertEqual(a(-10, -16).match(b).nonempty(), True)
        self.assertEqual(a(-20, -32).match(b).nonempty(), True)
        self.assertEqual(a(-40, -64).match(b).nonempty(), True)

    def test_components(self):

        pat = self.lt.pattern('''7b3o5b3o16b3o5b3o$6bo2bo5bo2bo15bo2bo3bo2bo$6bo3bo3bo3bo15bo9bo$7b4o3b
4o14b2o2bobobobo2b2o$8bobo3bobo15b2o2bobobobo2b2o$31bo5bo3bo5bo$10bo3b
o17bo2bobo3bobo2bo$5bo4b2ob2o4bo13b2o9b2o$5b2o3b2ob2o3b2o14bo9bo$10b2o
b2o$9b2o3b2o16b2o11b2o$9bo5bo16b2o11b2o2$36b2o3b2o$36b2o3b2o$19b3o8b3o
$3b3o3b2o3b2o3bo2bo7bo2bo12b3o$3bo2bo2b2o3b2o2bo3bo7bob2o12bo2bo$3bo
14bo3bo23bo2bo$4bobo11bo2bo$bo3bo13b2o12b3o5bob2o3bo2bo$bo25b2o4b2o7b
3o6bo$o26b3o2b4o10b2o2bo$29bo19bo$3bobo41b2o$3b3o$2bo2b3ob3o35bo$3b5o
3bo29b2o3bobo$2bobo2b2o2bo2bo26b3o2bobo$bobo5bo5bo$2bo7b2o4bo24b2obo$
3b2o6b3obo19bo6bo4b2o2bo$35bobo6bo6bo$35bo3b2o3bo$36bo4bo5b2o$36bo4bo$
37bo2bo!''', 'b3s23')

        ncomps = len(pat.components())
        self.assertEqual(ncomps, 41)

    def test_copy(self):

        pat = self.lt.pattern('xp40_33zxs48g69gvzx3021zy833')
        pat2 = pat.__copy__()

        self.assertEqual(pat, pat2)
        pat += pat(1093, 3511)
        self.assertNotEqual(pat, pat2)
        pat2 += pat2(1093, 3511)
        self.assertEqual(pat, pat2)

    def test_samples(self):

        pat = self.lt.pattern('xp40_33zxs48g69gvzx3021zy833')
        samples = pat.download_samples()
        periods = [s[10000][-1000:1000, -1000:1000].period for s in samples['C2_2']]
        self.assertEqual(periods[:20], [40]*20)

        synth = pat.download_synthesis()
        res = synth[1000]
        self.assertEqual(res.period, 40)

    def test_hashsoup(self):

        x = self.lt.hashsoup("b3s23", "C1", "n_GcpqZ62YqD3310086454")
        ac = x[1000].apgcode
        self.assertEqual(ac, 'xp46_033y133zzzckgsxsgkczydciczy4ccy0222')

    def test_lhistory(self):

        scstring = '''x = 50, y = 72, rule = LifeHistory
26.4B$25.4B$24.4B$19.3B.4B$18.8B10.E$18.7B.B7.3E$18.10B5.E$16.12B5.CE
$15.14B.5B$15.18B$14.22B$14.23B$12.ECB.22B$11.E.EB.23B$11.E4.22B$10.
2E5.20B$18.19B$17.21B$15.ECB.15B.4B$14.E.EB3.10B5.4B$14.E6.10B6.4B$
13.2E5.13B5.4B$19.15B5.4B$18.16B6.4B$17.17B7.4B$18.16B8.4B$19.13B11.
4B$19.5B2C2B.3B12.4B$21.3B2C2B2.4B11.4B$21.8B3.CE12.4B$20.8B4.E14.3B$
20.8B5.3E12.2B$20.7B8.E13.B$20.7B$21.6B7.E$21.6B6.E.E$22.5B6.E.E$22.
6B4.2E.3E$21.6B6.B4.E$21.7B3.B2CB3E$8.3B11.8B.B2C.E$7.6B9.10B$5.11B5.
3B2C6B$5.10B2E5.2B2C6B$5.11BE5.10B$5.11BEBEB2.11B$5.12BECB.12B$5.13B.
15B$5.11B3.16B$5.11B3.16B.2B$5.10B4.18BCE$5.6B7.17B.BCE$5.6B6.4B2.8B.
4B.B$5.4B7.4B4.7B$5.3B7.4B5.6B$6.B7.3DB6.4B$13.3BD5.E3B$12.3BD5.E.CB$
11.4B6.E.C$10.3AB8.E$9.3BA6.3E$8.3BA7.E$7.4B$6.4B$5.4B$4.4B$3.4B$2.4B
$.4B$4B$3B$2B!'''

        scorbie = self.lt.pattern(scstring)
        self.assertEqual(scorbie[2000].population, 4425)
        rlestring = scorbie.rle_string()

        scstring  =  scstring.replace('\n','').split('tory')[-1]
        rlestring = rlestring.replace('\n','').split('tory')[-1]

        self.assertEqual(scstring, rlestring)
 
    def test_stream(self):

        push44 = [0, 109, 90, 93, 91, 90, 95, 91, 91, 138, 157, 96, 90, 120, 91, 97, 107, 90, 90, 93, 188]
        one_glider = self.lt.pattern('3o$o$bo!')
        elbow = self.lt.pattern('2o$2o!')

        eg = one_glider(5, 2).stream(push44) + elbow
        self.assertEqual(eg[4096], elbow(-22, -22))

    def test_empty(self):

        empty_pattern = self.lt.pattern()
        self.assertEqual(empty_pattern.apgcode, 'xs0_0')

    def test_randfill(self):

        pat = self.lt.pattern()
        pat[70:200, 50:300] = 0.5
        self.assertEqual(pat.bounding_box, [70, 50, 130, 250])
        self.assertAlmostEqual(pat.population, 16250, delta=900) # 10 standard deviations

    def test_bbox(self):

        lidka = self.lt.pattern("bo$obo$bo8$8bo$6bobo$5b2obo2$4b3o!", 'b3s23')
        rect100 = [23 - (10 ** 100) // 4, 13 - (10 ** 100) // 4, (10 ** 100) // 2 - 206, (10 ** 100) // 2 - 186]
        self.assertEqual(lidka[10 ** 100].getrect(), rect100)

    def test_lidka(self):

        lidka = self.lt.pattern("bo$obo$bo8$8bo$6bobo$5b2obo2$4b3o!", 'b3s23')
        self.assertEqual(lidka.population, 13)
        lidka_30k = lidka[30000]
        self.assertEqual(lidka_30k.population, 1623)

        rect18 = [-249999999999999977, -249999999999999987, 499999999999999794, 499999999999999814]

        self.assertEqual(lidka[10 ** 18].getrect(), rect18)
        self.assertEqual(lidka[1093][3511], lidka[3511][1093])
        self.assertNotEqual(lidka[1093], lidka[3511])

    def test_coe(self):

        coe = self.lt.pattern('xq16_gcbgzvgg826frc')
        osc = coe.oscar()
        self.assertEqual(osc['period'], 16)

        self.assertEqual(coe.wechsler, 'gcbgzvgg826frc')

    def test_timeline(self):

        try:
            os.makedirs('lifelib/tempdir')
        except:
            pass

        try:
            lidka = self.lt.pattern("bo$obo$bo8$8bo$6bobo$5b2obo2$4b3o!")
            lframes = [lidka[i << 8] for i in range(120)]
            self.lt.save_timeline(lframes, 'lifelib/tempdir/timeline.mc')
            frames2 = self.lt.load('lifelib/tempdir/timeline.mc')

            self.assertEqual(len(lframes), len(frames2))
            for (x, y) in zip(lframes, frames2):
                self.assertEqual(x, y)

        finally:
            import shutil
            shutil.rmtree('lifelib/tempdir')

    def test_io(self):

        try:
            os.makedirs('lifelib/tempdir')
        except:
            pass

        try:
            lidka = self.lt.pattern("x = 9, y = 15, rule = B3/S23\nbo$obo$bo8$8bo$6bobo$5b2obo2$4b3o!")[30000]
            lidka.save('lifelib/tempdir/lidka.rle', tempfile='lifelib/tempdir/tempfile')
            lidka.save('lifelib/tempdir/lidka.rle.gz', tempfile='lifelib/tempdir/tempfile')
            lidka.save('lifelib/tempdir/lidka.mc', tempfile='lifelib/tempdir/tempfile')
            lidka.save('lifelib/tempdir/lidka.mc.gz', tempfile='lifelib/tempdir/tempfile')

            # Would be embarrassing if gzip didn't reduce file sizes:
            slr = os.path.getsize('lifelib/tempdir/lidka.rle')
            slrz = os.path.getsize('lifelib/tempdir/lidka.rle.gz')
            slm = os.path.getsize('lifelib/tempdir/lidka.mc')
            slmz = os.path.getsize('lifelib/tempdir/lidka.mc.gz')
            self.assertLess(slrz, slr)
            self.assertLess(slmz, slm)

            lr = self.lt.load('lifelib/tempdir/lidka.rle', tempfile='lifelib/tempdir/tempfile')
            lrz = self.lt.load('lifelib/tempdir/lidka.rle.gz', tempfile='lifelib/tempdir/tempfile')
            lm = self.lt.load('lifelib/tempdir/lidka.mc', tempfile='lifelib/tempdir/tempfile')
            lmz = self.lt.load('lifelib/tempdir/lidka.mc.gz', tempfile='lifelib/tempdir/tempfile')

            # Macrocell should respect overall position; RLE need not:
            self.assertEqual(lidka, lm)
            self.assertEqual(lidka.centre(), lr.centre())

            # Compression should make no difference:
            self.assertEqual(lr, lrz)
            self.assertEqual(lm, lmz)
        finally:
            import shutil
            shutil.rmtree('lifelib/tempdir')

    def make_p51(self):

        part = self.lt.pattern("3b2o3b5o$3b2o3b2ob2o$8b2o2bo$2o7bobo$2o8bo$10bo2bo$11bo2bo$11bobo!", "b3s23").shift(-15, -15)
        p51 = part + part('flip') + part('flip_x') + part('flip_y') + part('rot90') + part('rot270') + part('swap_xy') + part('swap_xy_flip')
        return p51

    def test_apgcode(self):

        desired_apgcode = 'xp51_oo033wgwgggy1gggwgw330oozy04e9130111y11110319e4zy077yf77zy0ooyfoozy09t40gy9g04t9z660gg01310333y133301310gg066zx11yh11'
        actual_apgcode = self.make_p51()[27].apgcode
        self.assertEqual(actual_apgcode, desired_apgcode)

    def test_large_apgcode(self):

        cordership = self.lt.pattern('''x = 113, y = 114, rule = B3/S23
79bo$78bobo$77b2ob2o$78bo2bo$80b3o$80bobo$81b2o$81b2o12b2o$95b2o3$79bo$78bobo
$78bobo$79bo$103b2o$103b2o3$87bo$86bobo$86bobo$65bobo19bo$64b2ob2o27bo14b2o$
65bob2o27b3o12b2o$69bo25b2o2bo$68bo28bo2bo$50bo17bo25bobo2bo$50bo44b3o$50bo8b
o37bo$57bobo35bobo$51b3o2bo3bo34b2o7bo$52bob5o45b2o$58bo44bobo$53b2o4$32b3o$
32bobo$32b3o19b2o$48bo5bo$46bobo3bo2bo$34b2o10bo3bo3bo$32b2ob2o10b7o$32b2o2b
2o$31b2o3b2o14bo$32b2o2b2o12bobo19bo$33bobo13bo2bo19b2o$50b2o19bobo3$25bo$25b
o28b2o$25bo8bo19b2o$32bobo9b2o16b2ob3o$26b3o2bo3bo8b2o18bo$27bob5o31bo3bo$33b
o32b2obo$28b2o37bo2bo$68b2o3$52b2o$52b2o4$36bo$35b4o$33b5obo$32bo6b2o$31bo3bo
b3o$25bo6bobo2b2o$13bobo7bob2o$12b2ob2o6bobo$13bob2o6b2o$2bo14bo$bobo12bo$2ob
2o11bo$bo2bo$3b3o$3bobo$4b2o$4b2o3$11b2o$2bo7bo2bo$3bo7b2o$3b2o$2b3o$36bo$34b
2o$3b3o29b2o$bo4bo12b2o$b2o3bo11bo2bo$3b4o12b2o7$6b2o$6b2o7$14b2o$14b2o!''')
        long_apgcode = cordership.apgcode
        self.assertEqual(len(long_apgcode), 323)

    def test_p51(self):

        p51 = self.make_p51()
        p51_period = p51.oscar(verbose=False)['period']
        self.assertEqual(p51_period, 51)
        self.assertEqual((p51 ** 7).population, (p51.population ** 7))
        p102_period = (p51 * p51).oscar(verbose=False)['period']
        self.assertEqual(p102_period, 102)

    def test_numpy(self):

        p51 = self.make_p51()

        try:
            import numpy as np
        except ImportError:
            return

        coords = np.array([[0, 1], [2, 3], [4, 5], [-6, 69], [73, -48]], dtype=np.int64)
        values = np.array([1, 0, 1, 1, 0], dtype=np.int64)

        p51[coords] = values
        values2 = p51[coords]

        self.assertEqual(np.all(values == values2), True)

        mat2pat = lambda M : self.lt.pattern('$'.join([''.join(['bo'[y] for y in x]) for x in M.tolist()]) + '!')

        A = np.array(np.random.uniform(0, 2, size=(34, 55)), dtype='uint64')
        B = np.array(np.random.uniform(0, 2, size=(55, 89)), dtype='uint64')
        C = np.dot(A, B) % 2

        pA = mat2pat(A)
        pB = mat2pat(B)
        pC = mat2pat(C)

        pAB = pA.__matmul__(pB)

        self.assertEqual(pC, pAB)

    def test_mutable(self):

        lidka = self.lt.pattern("bo$obo$bo8$8bo$6bobo$5b2obo2$4b3o!", 'b3s23')
        self.assertEqual(lidka.population, 13)
        x = lidka
        x += lidka(40, 0)
        self.assertEqual(lidka.population, 26)

    def test_spacefiller(self):

        sf = "20b3o3b3o$19bo2bo3bo2bo$4o18bo3bo18b4o$o3bo17bo3bo17bo3bo$o8bo12bo3bo"
        sf += "12bo8bo$bo2bo2b2o2bo25bo2b2o2bo2bo$6bo5bo7b3o3b3o7bo5bo$6bo5bo8bo5bo8b"
        sf += "o5bo$6bo5bo8b7o8bo5bo$bo2bo2b2o2bo2b2o4bo7bo4b2o2bo2b2o2bo2bo$o8bo3b2o"
        sf += "4b11o4b2o3bo8bo$o3bo9b2o17b2o9bo3bo$4o11b19o11b4o$16bobo11bobo$19b11o$"
        sf += "19bo9bo$20b9o$24bo$20b3o3b3o$22bo3bo2$21b3ob3o$21b3ob3o$20bob2ob2obo$"
        sf += "20b3o3b3o$21bo5bo!"

        spacefiller = self.lt.pattern(sf, "b3s23")
        self.assertEqual(spacefiller.population, 200)

        googol = 10 ** 100
        exppop = (googol // 2) ** 2 + 17 * (googol // 2) + 200

        sg = spacefiller[googol]

        self.assertEqual(sg.population, exppop)

        # We now take a huge rectangular region out of the spacefiller:

        left = -(3 ** 70)
        right = 4 ** 56
        top = 5 ** 20
        bottom = 7 ** 20

        hugerect = sg[left:right, top:bottom]

        self.assertEqual(hugerect.population, ((right - left) * (bottom - top)) // 2)

        # Ensure Boolean operations work correctly:

        self.assertEqual(sg.population, hugerect.population + (sg - hugerect).population)


    def test_oscar(self):

        sm  = "$12bo7bo$11bobo5bobo$10b2ob2o3b2ob2o$11bobo5bobo$9bob2o3bo3b2obo$9b7ob"
        sm += "7o$13b7o$12bob2ob2obo$9b2obo7bob2o$8b3ob2o5b2ob3o$7bo17bo$10bob2o5b2ob"
        sm += "o2$14bo3bo$9bob3o5b3obo$10bob3o3b3obo$14bo3bo$13b2o3b2o$12b2o5b2o$11b"
        sm += "3o5b3o$8b4o9b4o$7bo17bo$6b2o3b2o7b2o3b2o$7bobobobo5bobobobo$7b3o2b3o3b"
        sm += "3o2b3o$12bo7bo$7bo17bo2$7bobobo9bobobo$9b3o9b3o$6bob3o2bo5bo2b3obo$7b"
        sm += "2ob3o7b3ob2o$11b2o7b2o$11bob2o3b2obo$11bo3bobo3bo$3b3o5bo3bobo3bo5b3o$"
        sm += "3bob2o6bo5bo6b2obo$3bo2bo3b2ob2o3b2ob2o3bo2bo$6bo2b2o3bo3bo3b2o2bo$10b"
        sm += "o2bo5bo2bo$5bobobobo9bobobobo$3b2obo19bob2o$9bo13bo$6b2o17b2o2$5bo21bo"
        sm += "$4b2o21b2o$3b2o23b2o2$3b2o23b2o$4bo23bo$5bo4bobo7bobo4bo$4b3o4bob2o3b"
        sm += "2obo4b3o$4b3o2bobob2o3b2obobo2b3o$9bo2b2o5b2o2bo$8b4o9b4o2$6b2o17b2o$"
        sm += "6b2o2b2o9b2o2b2o$7bo2bobo7bobo2bo$4bo2bobo2bo3bo3bo2bobo2bo$4bo4bo3bob"
        sm += "3obo3bo4bo$5bobobo4bobobo4bobobo$7bobo4bo3bo4bobo$5b4o2bobo5bobo2b4o$"
        sm += "5b3ob2ob2ob3ob2ob2ob3o$14b5o$11bobo5bobo$11bob2o3b2obo$16bo$16bo$10b2o"
        sm += "9b2o$10b2obo5bob2o$10bob2obobob2obo$12b2obobob2o$9bobobobobobobobo$9b"
        sm += "2o2bo5bo2b2o$10bob2o5b2obo$12b2o5b2o$11bo9bo$11b3o5b3o4$12bo7bo$11b2ob"
        sm += "o3bob2o$11bo2bo3bo2bo$11b3o5b3o$13b3ob3o$14b2ob2o$15bobo$14b2ob2o$12b"
        sm += "2o2bo2b2o$16bo$11b2o7b2o$10b3obo3bob3o$9bo2bo7bo2bo$10bo11bo$11bo9bo$"
        sm += "8b2ob2o7b2ob2o$11bo9bo$7bo3b2o7b2o3bo$8b6o5b6o$12bo7bo$11b2obo3bob2o$"
        sm += "12bob2ob2obo$11bo9bo$12bobo3bobo$10bobobo3bobobo$12bob5obo$11b2ob2ob2o"
        sm += "b2o$11b4o3b4o$10bo2bo5bo2bo$9bo2bo7bo2bo$10bob2o5b2obo$8bob2o9b2obo$8b"
        sm += "2o13b2o$8b2o13b2o$9b2o2b3ob3o2b2o$9b2o2b2o3b2o2b2o$12b2o5b2o$9bo2bo7bo"
        sm += "2bo$9bo2bo7bo2bo$8b2obo9bob2o$8bo2bo3b3o3bo2bo$8bo2bo3b3o3bo2bo$11bo2b"
        sm += "o3bo2bo$7b2obo4b3o4bob2o$6b6o3bobo3b6o$6bo19bo$10bo2b3ob3o2bo$10bo3bo"
        sm += "3bo3bo$10bo11bo$15b3o$16bo$13b3ob3o$14b2ob2o!"

        monster = self.lt.pattern(sm, "b3s23")

        results = monster.oscar(verbose=False)

        self.assertEqual(results['period'], 7)
        self.assertEqual(results['displacement'], (0, -3))

        galaxy = self.lt.pattern("6b2o$2o5bo$obo3bo$3bobo$4bo$3bobo$2bo3bobo$bo5b2o$b2o!", "b3s23")

        results = galaxy.oscar(verbose=False)

        self.assertEqual(results['period'], 8)

        results = galaxy.oscar(verbose=False, eventual_oscillator=False)

        self.assertEqual(len(results), 0)

    def test_p57(self):

        just_eater = self.lt.pattern('''8b3o$5b2obobo$5bo2bobo$3b3o$3b2o$4bo26b2o$32bo$29b3o$29bo$3bo3b2o$2bob
        o3bo$bobo3bo12b2o$bo4bo13bobo$2o5b3o12bo$9bo12b2o!''', 'b3s23')

        with_catalyst = self.lt.pattern('''8b3o$5b2obobo$5bo2bobo$3b3o$3b2o$4bo26b2o$32bo$29b3o$22bo6bo$3bo3b2o
        14bo$2bobo3bo12b3o$bobo3bo6b2o$bo5b2o5b3o$2o5b4ob2obob2o$o3b2o4b3o4b3o
        $b4obo6b2o6bo$10bo2b2ob6o$b2ob2o2bobo8bo2b2o$2bobo2b2o3b3o2bo2bo3bo$bo
        3b3obo8bo2b3o$2b3o4bo4bo3bo4bobo$5bo4b2obo10b2o$4bo2b3o2b2ob6o$5b2o3bo
        3bobobobo14bobo$7bobob2obo3bo17b2o$7bobo2bo6bobo14bo$8bobo2b2o5bobo$9b
        o5b2ob2ob2o$15b3o2bo2b2o$20bo2bo$13bob5obobo$13b2o2bobob2o$11b2o4bo2bo
        $10bo2b4ob2o2b3o$10b2o5bo2b2o2bo$15bo3bo2bo$15b2o3b2o$51bo$49bobo$50b
        2o!''', 'b3s23')

        p57loop = self.lt.pattern('''14bob2o9bo35b2o12bo35b2o12bo35b2o12bo35b2o12bo35b2o12bo35b2o12bo35b2o
        12bo35b2o12bo35b2o12bo35b2o12bo$14b2obo9b3o5b2o27bo12b3o5b2o27bo12b3o
        5b2o27bo12b3o5b2o27bo12b3o5b2o27bo12b3o5b2o27bo12b3o5b2o27bo12b3o5b2o
        27bo12b3o5b2o27bo12b3o5b2o27bo12b3o5b2o$30bo4bo28bobo13bo4bo28bobo13bo
        4bo28bobo13bo4bo28bobo13bo4bo28bobo13bo4bo28bobo13bo4bo28bobo13bo4bo
        28bobo13bo4bo28bobo13bo4bo28bobo13bo4bo$12b5o12bo3bobo29b2o12bo3bobo
        29b2o12bo3bobo29b2o12bo3bobo29b2o12bo3bobo29b2o12bo3bobo29b2o12bo3bobo
        29b2o12bo3bobo29b2o12bo3bobo29b2o12bo3bobo29b2o12bo3bobo$11bo4bo11bo3b
        obo43bo3bobo43bo3bobo43bo3bobo43bo3bobo43bo3bobo34bo8bo3bobo43bo3bobo
        43bo3bobo34b2o7bo3bobo43bo3bobo$10bo2bo14b2o3bo44b2o3bo44b2o3bo44b2o3b
        o44b2o3bo44b2o3bo34b3o7b2o3bo44b2o2bo45b2o3bo34bo9b2o3bo44b2o3bo$7bo2b
        ob2o43bo49bo11b2o36bo11bo37bo49bo49bo9bo39bo24bo24bo49bo11b2o36bo49bo
        16bo$6bobobo5bo21b2o15b3o47b3o9bobo35b3o10b3o15bo18b3o47b3o47b3o8bo2b
        2o34b3o47b3o47b3o11bo35b3o32bo14b3o8b2o5bobo$7bo2bo4b2o19b5o13bo49bo
        13bo35bo12bo19bo3bo12bo27b2o20bo49bo29b2o18bo30bo3bo14bo49bo13bo10bo
        24bo31b2o2bo13bo10bo2bo5bo$10b2o2b3o18bo4bo13b2o48b2o21bob3o9b2o11b2o
        10b2ob3o10b2o3bo4bo11b2o25bobo7b3o10b2o26bo8b2o11b2o10bob2ob3o6b2ob3o
        5b2o11b2o28bo3b2o14b2o48b2o12bo9bob2o9b2o11b2o12b3o14bo5bo12b2o7bo2bob
        o$35bo2b3o39b3o9b2o31b2obo3bo8b2o7bo16b2o3bo9bo3bo2bo2bo40bo6b2ob2o37b
        2o7b2o24b2o2bo2bo4b3ob2o6b2o44b2o52b2o8bo25bo4bo8b2o24bo2bo13b2o2bobo
        21bo4bo4b5o$21bo13b2o2bo42bo8b2o33bo4b2o16b4o18bo10bobo3b4o30bo17bo2b
        2o36b3o12bo5bo16bo3bo6b2ob3o52bo41bo17bo29bo42bo13b2o3bo21bo9bo4bo$20b
        2o15b2o35b4o3bo10bo34b4obo14b2o3b2o15bobo11b2o3b2o31b3o17b2o33bobo2bo
        12bo3bo2b2o15bo14b2o33b3o18bo38b2ob2o8b2obob3o30b4o17b3o13bo4bo14b3o
        25b3o8bo2bo$8b2o7bo55b2o71bo3bo69bobobob2o48bobob2o12bo4b3o17bobo29b3o
        13bo2bo2bo17bobo31b5o11b2ob5o3bo16bo29b2obo13bo58b2obo$7bo2bo4bo2bo53b
        2obo2bo67bo3bo68b2o3bob2o48b3o14bo5b2o19bo29bo4bo11bo4bo20bobo29bo2bo
        11bo4b2o4bo17bo28b2obo70bo5bob2o$7bobobo3bo38bobo16bobob2o33bo33b2o2bo
        50b2o3bo13b2obo2bo32bo34b3o2b2o15b3o31b2obo13bobo2b3o50bo2bo12bobo55bo
        bo17bo51bobo4bobo$8bo2b3o2bo2bo3b2o28bo3bo16b2obo32b4o34bo3bo13b2o33b
        2ob2ob2o13b2ob2obo15b3o10b2o35b2o58b3o11b2o6b3o14bo12bo19b2o15bob2o15b
        2o36b2o4bo12bo33bo2b2o13bo2bo2b2o2bo$17b3o32bo4bo17b2o17b3o12bob5o40bo
        8bo2bo32bo7bo12b4o19bo10b2o38bo17b2o41bo18b3o14bo10bobo36b3o15bo2bo39b
        obo9b3o33b2o2b3o13b2o6b2o$10b2o40bo4bo8b2o25bo14b2o6bo40b2o7b3o31b2o6b
        o7b2o6b3o3bo13b2ob2o6bo11bo46bo4bo46b2o27bobo6b3o9b2o26bo16bo2bo40b2o
        44b3o2b2o$10bo18b2o21bo3bo9b2o11b2o12bo2bo12b3obo3bo11b2o25b2o8b2o11b
        2o20b2o2b2o9b2o12bo9bobo2b3o8b5o5b2o11b2o26bobo4bo14b2o35b2o11b2o11bo
        5b2o4bo11b2o11b2o30b2ob2o13b2o48b2o21b2o2b2o$29bo13bobo7bo2bo22bo13b3o
        14b2o2bobo12bo49bo21b2o2b2o22bo11bo4bo10bo21bo28b2o5bo2bo10bo49bo12bob
        o34bo33b2o14bo49bo$30b3o9bo37b3o47b3o47b3o12bo34b3o9bo2b2o33b3o25bo8bo
        12b3o47b3o47b3o47b3o47b3o9b3o$19b2o3b2ob3o2bo9bo2bo36bo49bo49bo12bo36b
        o11b2o36bo49bo49bo11bo37bo24bo24bo49bo9b3o$20b2obobobo2bo12bobo7b2o3bo
        44b2o3bo44b2o3bo35bo8b2o3bo35bo8b2o3bo44b2o3bo44b2o3bo33bo2bo7b2o3bo
        44b2o3bo44b2o3bo35bo8b2o3bo$6b2o11bo4bo2bobo23bo3bobo43bo3bobo43bo3bob
        o43bo3bobo43bo3bobo43bo3bobo43bo3bobo35bo7bo3bobo43bo5bo43bo3bobo43bo
        3bobo$7bo19bob2o9b2o12bo3bobo29b2o12bo3bobo29b2o12bo3bobo29b2o12bo3bob
        o29b2o12bo3bobo29b2o12bo3bobo29b2o12bo3bobo29bo3bo9bo3bobo29b2o12bo2b
        2obo29b2o12bo3bobo29b2o12bo3b2o$7bobo16b2o3bo7bobo13bo4bo28bobo13bo4bo
        28bobo13bo4bo28bobo13bo4bo28bobo13bo4bo28bobo13bo4bo28bobo13bo4bo28bob
        2o12bo4bo28bobo13bo4bo28bobo13bo4bo28bobo13bo11b3o$8b2o13bobo2bobo8bo
        12b3o5b2o27bo12b3o5b2o27bo12b3o5b2o27bo12b3o5b2o27bo12b3o5b2o27bo12b3o
        5b2o27bo12b3o5b2o27bo12b3o5b2o27bo12b3o5b2o27bo12b3o5b2o27bo13bob5o7bo
        12b2o$23b2o2b2ob2o6b2o12bo35b2o12bo35b2o12bo35b2o12bo35b2o12bo35b2o12b
        o35b2o12bo35b2o12bo35b2o12bo35b2o12bo35b2o13b2o4bo7b3o6b2o2bo$11b2o
        543b2obob2o13bobobo$12b2o539b2obobobobo5bo10bo$13b3o537bobo4bo7b3o$13b
        3o543b2o9b3o4bo$12bo557b2o4bobo$577bo$11bo3bobo560b3o$15b2o563bo$12b2o
        b2o$14bo538bo$9b2o542b3o$9b2o14b2o529bo$25bo529b2o14b2o$26b3o542b2o$
        28bo538bo$565b2ob2o$bo563b2o$b3o560bobo3bo$4bo$3bobo4b2o557bo$4bo4b3o
        9b2o543b3o$11b3o7bo4bobo537b3o$2bo10bo5bobobobob2o539b2o$bobobo13b2obo
        b2o543b2o$bo2b2o6b3o7bo4b2o13b2o35bo12b2o35bo12b2o35bo12b2o35bo12b2o
        35bo12b2o35bo12b2o35bo12b2o35bo12b2o35bo12b2o35bo12b2o6b2ob2o2b2o$2o
        12bo7b5obo13bo27b2o5b3o12bo27b2o5b3o12bo27b2o5b3o12bo27b2o5b3o12bo27b
        2o5b3o12bo27b2o5b3o12bo27b2o5b3o12bo27b2o5b3o12bo27b2o5b3o12bo27b2o5b
        3o12bo8bobo2bobo13b2o$12b3o11bo13bobo28bo4bo13bobo28bo4bo13bobo28bo4bo
        12b2obo28bo4bo13bobo28bo4bo13bobo28bo4bo13bobo28bo4bo13bobo28bo4bo13bo
        bo28bo4bo13bobo28bo4bo13bobo7bo3b2o16bobo$22b2o3bo12b2o29bobo3bo12b2o
        29bob2o2bo12b2o29bobo3bo9bo3bo29bobo3bo12b2o29bobo3bo12b2o29bobo3bo12b
        2o29bobo3bo12b2o29bobo3bo12b2o29bobo3bo12b2o29bobo3bo12b2o9b2obo19bo$
        22bobo3bo43bobo3bo43bo5bo43bobo3bo7bo35bobo3bo43bobo3bo43bobo3bo43bobo
        3bo43bobo3bo43bobo3bo43bobo3bo23bobo2bo4bo11b2o$23bo3b2o8bo35bo3b2o44b
        o3b2o44bo3b2o7bo2bo33bo3b2o44bo3b2o44bo3b2o8bo35bo3b2o8bo35bo3b2o44bo
        3b2o44bo3b2o7bobo12bo2bobobob2o$37b3o9bo49bo24bo24bo37bo11bo49bo49bo
        36b2o11bo36bo12bo49bo49bo36bo2bo9bo2b3ob2o3b2o$37b3o9b3o47b3o47b3o47b
        3o47b3o12bo8bo25b3o33b2o2bo9b3o34bo12b3o47b3o47b3o37bo9b3o$52bo49bo14b
        2o33bo34bobo12bo49bo10bo2bo5b2o28bo21bo10bo4bo11bo22b2o2b2o21bo49bo12b
        obo2b2o14b3o13bo22bo2bo7bobo13bo$24b2o2b2o21b2o48b2o13b2ob2o30b2o11b2o
        11bo4b2o5bo11b2o11b2o35b2o14bo4bobo26b2o11b2o5b5o8b3o2bobo9bo12b2o9b2o
        2b2o20b2o11b2o8b2o25b2o11bo3bob3o12bo2bo12b2o11b2o9bo3bo21b2o18bo$24b
        2o2b3o44b2o40bo2bo16bo26b2o9b3o6bobo27b2o46bo4bo46bo11bo6b2ob2o13bo3b
        3o6b2o7bo6b2o31b3o7b2o40bo6b2o14bo25b2o8bo4bo40b2o$2o6b2o13b3o2b2o33b
        3o9bobo39bo2bo15b3o36bobo10bo14b3o18bo41b2o17bo38b2o10bo19b4o12bo7bo
        32bo2bo8bo40b5obo12b3o17b2o17bo4bo32b3o$o2b2o2bo2bo13b2o2bo33bo12bo4b
        2o36b2o15b2obo15b2o19bo12bo14b3o6b2o11b3o58b2o35b2o10b3o15bob2ob2o13b
        2ob2ob2o33b2o13bo3bo34b4o32bob2o16bo3bo28b2o3bo2bo2b3o2bo$bobo4bobo51b
        o17bobo55bobo12bo2bo50b3o2bobo13bob2o31b3o15b2o2b3o34bo32bo2bob2o13bo
        3b2o50bo2b2o33bo33b2obobo16bobo38bo3bobobo$2obo5bo70bob2o28bo17bo4b2o
        4bo11bo2bo29bobo20bo4bo11bo4bo29bo19b2o5bo14b3o48b2obo3b2o68bo3bo67bo
        2bob2o53bo2bo4bo2bo$3bob2o58bo13bob2o29bo16bo3b5ob2o11b5o31bobo17bo2bo
        2bo13b3o29bobo17b3o4bo12b2obobo48b2obobobo69bo3bo71b2o55bo7b2o$3bo2bo
        8b3o25b3o14bo4bo13b3o17b4o30b3obob2o8b2ob2o38bo18b3o33b2o14bo15b2o2bo
        3bo12bo2bobo33b2o17b3o31b2o3b2o11bobo15b2o3b2o14bob4o34bo10bo3b4o35b2o
        15b2o$4bo4bo9bo21bo3b2o13bo42bo29bo17bo41bo52b3ob2o6bo3bo16bo5bo12b3o
        36b2o2bo17bo30b4o3bobo10bo18b4o16b2o4bo33b2o8bo42bo2b2o13bo$5b5o4bo4bo
        21bobo2b2o13bo2bo24b2o8bo4bo25bo8b2o52b2o44b2o6b2ob3o4bo2bo2b2o24b2o7b
        2o37b2ob2o6bo40bo2bo2bo3bo9bo3b2o16bo7b2o8bo3bob2o31b2o9b3o39b3o2bo$
        13bobo2bo7b2o12bo5bo14b3o12b2o11b2o9b2obo9bo12b2o48b2o14b2o3bo28b2o11b
        2o5b3ob2o6b3ob2obo10b2o11b2o8bo26b2o10b3o7bobo25b2o11bo4bo3b2o10b3ob2o
        10b2o11b2o9b3obo21b2o48b2o13bo4bo18b3o2b2o$7bo5bo2bo10bo13bo2b2o31bo
        24bo10bo13bo49bo14bo3bo30bo18b2o29bo49bo20b2o27bo12bo3bo19bo12bo35bo
        13bo49bo13b5o19b2o4bo2bo$6bobo5b2o8b3o14bo32b3o35bo11b3o47b3o47b3o34b
        2o2bo8b3o47b3o47b3o18bo15b3o10b3o35bobo9b3o47b3o15b2o21bo5bobobo$7bo
        16bo49bo36b2o11bo49bo24bo24bo39bo9bo49bo49bo37bo11bo36b2o11bo49bo43b2o
        bo2bo$48bo3b2o44bo3b2o9bo34bo3b2o45bo2b2o44bo3b2o7b3o34bo3b2o44bo3b2o
        44bo3b2o44bo3b2o44bo3b2o44bo3b2o14bo2bo$47bobo3bo43bobo3bo7b2o34bobo3b
        o43bobo3bo43bobo3bo8bo34bobo3bo43bobo3bo43bobo3bo43bobo3bo43bobo3bo43b
        obo3bo11bo4bo$46bobo3bo12b2o29bobo3bo12b2o29bobo3bo12b2o29bobo3bo12b2o
        29bobo3bo12b2o29bobo3bo12b2o29bobo3bo12b2o29bobo3bo12b2o29bobo3bo12b2o
        29bobo3bo12b2o29bobo3bo12b5o$46bo4bo13bobo28bo4bo13bobo28bo4bo13bobo
        28bo4bo13bobo28bo4bo13bobo28bo4bo13bobo28bo4bo13bobo28bo4bo13bobo28bo
        4bo13bobo28bo4bo13bobo28bo4bo$45b2o5b3o12bo27b2o5b3o12bo27b2o5b3o12bo
        27b2o5b3o12bo27b2o5b3o12bo27b2o5b3o12bo27b2o5b3o12bo27b2o5b3o12bo27b2o
        5b3o12bo27b2o5b3o12bo27b2o5b3o12bo$54bo12b2o35bo12b2o35bo12b2o35bo12b
        2o35bo12b2o35bo12b2o35bo12b2o35bo12b2o35bo12b2o35bo12b2o35bo11bobo$
        567bo!''', 'b3s23')

        self.assertEqual(p57loop.period, 57)

        p57a = p57loop.apgcode
        self.assertEqual(len(p57a), 3083)
        self.assertEqual(self.lt.pattern(p57a).period, 57)

        p57gun = p57loop.replace(just_eater, with_catalyst, n_phases=57, orientations=['identity', 'rot180'])

        self.assertEqual(p57gun.population, 6690)
        self.assertEqual(p57gun[5700].population, 16690)

if __name__ == '__main__':
    unittest.main()




import os
from .rulefiles import rules_dir

def syms(rulestring):

    filename = rulestring if (rulestring[0] == 'x') else ('x' + rulestring)
    with open(os.path.join(rules_dir, filename + '.h')) as f:
        for line in f:
            if 'SYMS="' in line:
                return line.split('"')[1]

    return "NONE"

def zoi(rulestring):

    filename = rulestring if (rulestring[0] == 'x') else ('x' + rulestring)
    with open(os.path.join(rules_dir, filename + '.h')) as f:
        for line in f:
            if 'ZOI="' in line:
                return line.split('"')[1]

    return "99"

def bitplanes(rulestring):

    filename = rulestring if (rulestring[0] == 'x') else ('x' + rulestring)
    with open(os.path.join(rules_dir, filename + '.h')) as f:
        for line in f:
            break

    if '16' in line:
        return 16
    else:
        return 8

def family(rulestring):

    bp = bitplanes(rulestring)
    log2bp = len(bin(bp)) - 3
    return 2 * log2bp

mantissa = {0, 1}

def create_rule(rulestring):

    with open('iterators_%s.h' % rulestring, 'w') as f:
        f.write('#pragma once\n')
        filename = rulestring if (rulestring[0] == 'x') else ('x' + rulestring)
        f.write('#include "../../rules/%s.h"\n' % filename)




family = 4
bitplanes = 1
mantissa = {0, 1}

from .exceptions import NonCanonicalError

def create_rule(rulestring):
    isotrope(rulestring)

    logstring = rulestring[rulestring.index('b'):]

    with open('iterators_%s.h' % rulestring, 'w') as f:
        f.write('#pragma once\n')
        f.write('#include <stdint.h>\n')
        f.write('#include <cstdlib>\n')
        f.write('#include <iostream>\n')
        f.write('#include "../eors.h"\n')
        f.write('#include "../lifeperm.h"\n')
        f.write('namespace %s {\n\n' % rulestring.replace('-', '_'))
        f.write('#include "ma_%s.h"\n\n' % rulestring)
        f.write('#include "../isoluts.h"\n')
        f.write('}\n')


def squarelord():

    lord = ""
    lord += "_ceaccaieaeaknja_ceaccaieaeaknjaekejanaairerririekejanaairerriri";
    lord += "ccknncqnaijaqnwaccknncqnaijaqnwakykkqyqjrtjnzrqakykkqyqjrtjnzrqa";
    lord += "ekirkyrtejerkkjnekirkyrtejerkkjnekejjkrnejecjyccekejjkrnejecjycc";
    lord += "anriqyzraariqjqaanriqyzraariqjqajkjywkqkrnccqkncjkjywkqkrnccqknc";
    lord += "cnkqccnnkqkqyykjcnkqccnnkqkqyykjaqjwinaarzjqtrnaaqjwinaarzjqtrna";
    lord += "ccyyccyennkjyekeccyyccyennkjyekenykknejeirykrikenykknejeirykrike";
    lord += "aqrznyirjwjqkkykaqrznyirjwjqkkykaqrqajiarqcnnkccaqrqajiarqcnnkcc";
    lord += "intrneriaanajekeintrneriaanajekeajnkaeaeiaccaec_ajnkaeaeiaccaec_";
    return lord


def hexlord():

    lord = ""
    prelord = "___o_moo" + "_pmmomoo" + "_mpmmpmm" + "ommpomo_";
    prelord += prelord[::-1];
    for i in range(512):
        j = (i & 3) + ((i & 32) >> 3) + ((i & 256) >> 5) + ((i & 128) >> 3) + ((i & 8) << 2);
        lord += prelord[j]
    return lord


def tab2str(lord2, hexagonal):

    rule_letters = {}

    if hexagonal:
        lord = hexlord()
        rule_letters[2] = "omp"
        rule_letters[3] = "omp"
        rule_letters[4] = "omp"
    else:
        lord = squarelord()
        rule_letters[1] = "ce"
        rule_letters[2] = "ceaikn"
        rule_letters[3] = "ceaiknjqry"
        rule_letters[4] = "ceaiknjqrytwz"
        rule_letters[5] = "ceaiknjqry"
        rule_letters[6] = "ceaikn"
        rule_letters[7] = "ce"

    canonicals = [['' for i in range(9)] for j in range(2)]
    canstring = ''

    for i in range(512):
        centre = (i >> 4) & 1
        ncount = masked_popcount(i, hexagonal)
        if (lord2[i] == 1):
            canonicals[centre][ncount] += lord[i]

    for (centre, z) in enumerate('bs'):
        canstring += z
        for ncount in range(9):
            goodies = set(canonicals[centre][ncount])
            if (len(goodies) == 0):
                continue
            canstring += str(ncount)
            if ncount not in rule_letters:
                continue
            a = ''.join(sorted([c for c in rule_letters[ncount] if c in goodies]))
            b = '-' + (''.join(sorted([c for c in rule_letters[ncount] if c not in goodies])))
            if (len(b) == 1):
                continue
            elif (len(b) < len(a)):
                canstring += b
            elif (len(a) < len(b)):
                canstring += a
            else:
                canstring += (b if hexagonal else a)

    if hexagonal:
        canstring += 'h'

    return canstring


def masked_popcount(i, hexagonal):

    popcounts = [0, 1, 1, 2, 1, 2, 2, 3, 1, 2, 2, 3, 2, 3, 3, 4]

    if hexagonal:
        i = i & 443

    return popcounts[i & 15] + popcounts[i >> 5]


def str2tab(rulestring, hexagonal='unknown'):

    if (hexagonal == 'unknown'):
        hexagonal = (rulestring[-1].lower() == 'h')

    isotrans = {}
    centre = 0

    for c in rulestring.lower().replace('v', 'r'):
        if c in '012345678':
            lastloc = int(c)
            isotrans[(centre, lastloc)] = "+";
        elif (c == 'b'):
            centre = 0
        elif (c == 's'):
            centre = 1
        elif (c == '/'):
            centre = 1 - centre
        elif c in ('omp-' if hexagonal else 'ceaiknjqrytwz-'):
            isotrans[(centre, lastloc)] += c

    if not hexagonal:

        lord = squarelord()

    else:

        lord = hexlord()

        import re

        regex = '1?(2([-]?[omp])?)?(3([-]?[omp])?)?(4([-]?[omp])?)?5?6?'
        regex = '^b' + regex + 's0?' + regex + 'h$'

        if not re.match(regex, rulestring):

            raise ValueError('Error: "%s" does not match the regex "%s"' % (rulestring, regex))

    lord2 = []

    for i in range(512):

        centre = (i >> 4) & 1

        ncount = masked_popcount(i, hexagonal)

        if (centre, ncount) in isotrans:
            fragment = isotrans[(centre, ncount)]
            if (len(fragment) == 1):
                lord2.append(1)
            else:
                l = 1 if (fragment[1] == '-') else 0
                for c in fragment:
                    if (c == lord[i]):
                        l = 1 - l
                lord2.append(l)
        else:
            lord2.append(0)

    canstring = tab2str(lord2, hexagonal)

    if (rulestring != canstring):
        raise NonCanonicalError('%s is a non-canonical version of %s' % (rulestring, canstring))

    return lord2


def isotrope(rulestring):

    lord2 = str2tab(rulestring)

    lut9 = [sum([(lord2[8*i+j] << j) for j in range(8)]) for i in range(64)]

    def oi(i0, i1, i2, i4, i5, i6):
        thisbyte = 0
        for i3 in range(2):
            for i7 in range(2):
                for i8 in range(2):
                    thisindex = sum([(j << i) for (i, j) in enumerate([i0, i1, i2, i3, i4, i5, i6, i7, i8])])
                    thisbit = 1 - lord2[thisindex]
                    thisbyte |= (thisbit << (4 * i3 + 2 * i7 + i8))
        return thisbyte

    def ora(i4, i5):
        return [oi(i0, i1, i2, i4, i5, i6) for i6 in range(2) for i1 in range(2) for i2 in range(2) for i0 in range(2)]

    origarrays = [ora(i4, i5) for i5 in range(2) for i4 in range(2)]

    maskarrays = [[0 for i in range(16)] for j in range(8)]

    maskarrays[0] = [a ^ b for (a, b) in zip(origarrays[0], origarrays[2])]
    maskarrays[1] = [a ^ b for (a, b) in zip(origarrays[1], origarrays[3])]
    maskarrays[2] = origarrays[2]
    maskarrays[3] = origarrays[3]
    maskarrays[4] = [j for i in range(8) for j in [2 ** i, 0]]
    maskarrays[5] = [j for i in range(8) for j in [0, 2 ** i]]
    maskarrays[6] = [127] * 16
    maskarrays[7] = [112] * 16

    lut9 += [0, 4, 8, 12, 1, 5, 9, 13, 2, 6, 10, 14, 3, 7, 11, 15]
    lut9 += [255, 0, 255, 0, 255, 0, 255, 0, 255, 0, 255, 0, 255, 0, 255, 0]
    lut9 += [255, 255, 0, 0, 255, 255, 0, 0, 255, 255, 0, 0, 255, 255, 0, 0]

    with open('ma_%s.h' % rulestring, 'w') as f:

        # Double every mask array for avx2 compatibility:
        marray = [x for m in maskarrays for x in (m + m)]
        f.write('// Two hundred and fifty-six seemingly arbitrary bytes\n')
        f.write('const static uint8_t __magicsauce[] __attribute__((aligned(64))) = {%d,\n' % marray[0])
        for i in range(15):
            currstring = '   '
            for j in range(i*17+1, i*17+18):
                currstring += ((' %3d};' if (j == 255) else ' %3d,') % marray[j])
            f.write(currstring + '\n')

        f.write('\n// 512-bit lookup table with extra things appended\n')
        f.write('const static uint8_t lut9[] __attribute__((aligned(64))) = {%d, %d, %d, %d,\n' % tuple(lut9[:4]))
        for i in range(6):
            currstring = '   '
            for j in range(i*18+4, i*18+22):
                currstring += ((' %3d};' if (j == 111) else ' %3d,') % lut9[j])
            f.write(currstring + '\n')



def determine_if_gun(a, maxexp):

    lt4 = a.owner

    if lt4.n_layers < 4:
        lt4 = lt4.session.lifetree(n_layers=4)

    a = a[524288]
    x = lt4.pattern('', 'LifeHistory')
    x += a
    x = x[524288]
    oldpop = x.population
    x = x[524288]
    newpop = x.population
    p = newpop - oldpop

    if (p < 500000):
        return None

    if (p > 550000):
        return None

    envelope = x.layers()[1]
    livecells = x[8].layers()[0]
    y = livecells - envelope
    envelope = x[256].layers()[1]

    if (y.population != 5):
        return None

    z = lt4.pattern("", "b3s23")
    z += y

    osc = z.oscar(maxexp=maxexp, return_apgcode=True, eventual_oscillator=False, verbose=False, allow_guns=False)

    if osc.get('apgcode', '') != 'xq4_153':
        return None

    w = lt4.pattern("", "LifeHistory")
    w += z[-2097152]
    w = w[4194304]


    band = w.layers()[1]

    pbb = envelope - band
    r1 = z.getrect()
    r2 = z[1024].getrect()

    pbb += pbb(r1[0] - r2[0], r1[1] - r2[1])

    r = pbb.getrect()
    if (len(r) != 4):
        return None

    envelope = envelope[r[0] : r[0] + r[2], r[1] : r[1] + r[3]]
    r = envelope.getrect()

    geater = lt4.pattern("bo$2bo$3o4$5b2o$5bo$6b3o$8bo!", "b3s23")

    eater = z.replace(geater[:3, :3], geater[100], orientations='rotate4reflect', n_phases=2)

    if (eater.population != 7):
        return None

    x += eater

    x = x[128]

    osc = x.oscar(maxexp=maxexp, eventual_oscillator=False, allow_guns=False, verbose=False)

    if 'period' not in osc:
        return None

    multiple = osc['period']
    gun = lt4.pattern("", "b3s23")
    gun += livecells

    if ((gun - gun[multiple]).population > 0):
        return None

    lastshadow = None
    canonical_gun = None
    cw = None

    for i in range(multiple + 1):

        gun = gun[1]
        nextshadow = gun & envelope
        admissible = (i > 0) and (lastshadow[1] == nextshadow)
        lastshadow = nextshadow

        if admissible and ((nextshadow[1] - envelope).population > 0):
            w = nextshadow.phase_wechsler()
            if (cw is None) or (len(cw) > len(w)) or ((len(cw) == len(w)) and (cw > w)):
                cw = w
                canonical_gun = nextshadow

    if cw is None:
        return None

    return (canonical_gun, multiple, cw, z, envelope)





import unittest
import lifelib
import os

class TestNutshell(unittest.TestCase):

    def test_brew(self):

        try:
            import nutshell
            print("Nutshell module successfully loaded.")
        except ImportError:
            print("Nutshell module not found.")
            return

        filename = os.path.join(lifelib.lifelib_dir, 'rules', 'source', 'Brew.ruel')
        sess = lifelib.load_rules(filename, force_compile=True)
        lt = sess.lifetree()
        p150 = lt.pattern('''x = 30, y = 21, rule = Brew
11.2B4.2B$10.B2.B2.B2.B$11.2B4.2B2$13.4C$2.C9.C4.C9.C$3.2C4.BCB6.BCB
4.2C$2C.C5.BCBC4.CBCB5.C.2C$2.2C3.2B.2A6.2A.2B3.2C$.C4.C.AB2CBA2.AB2C
BA.C4.C$.C3.2BCA4BA2.A4BAC2B3.C$.C4.C.AB2CBA2.AB2CBA.C4.C$2.2C3.2B.2A
6.2A.2B3.2C$2C.C5.BCBC4.CBCB5.C.2C$3.2C4.BCB6.BCB4.2C$2.C9.C4.C9.C$
13.4C2$11.2B4.2B$10.B2.B2.B2.B$11.2B4.2B!''')

        self.assertEqual(p150.period, 150)
        self.assertEqual(p150.apgcode, 'xp150_y3coa4aoczy4vxvzy33152513_o4oy0cen9necy0o4ozg9gy2vfvy2g9gz121y037e9e73y0121')



from .parsetable import GetNumberOfInputs

class RuleTree:
    '''
    Based on the RuleTree code from Golly's glife module, largely
    written by Tim Hutton, Tom Rokicki, and Andrew Trevorrow.
    '''

    def __init__(self, numStates, numInputs, initialise=True):

        self.numInputs = numInputs;

        self.world = {} # dictionary mapping node tuples to node index (for speedy access by value)
        self.seq = [] # same node tuples but stored in a list (for access by index)
        # each node tuple is ( depth, index0, index1, .. index(numStates-1) )
        # where each index is an index into self.seq

        self.nodeSeq = 0
        self.curndd = -1
        self.numStates = numStates

        self.cache = {}
        self.shrinksize = 100

        if initialise:
            self._init_tree()

    def _init_tree(self):
        self.curndd = -1
        for i in range(self.numInputs):
            node = tuple( [i+1] + [self.curndd]*self.numStates )
            self.curndd = self._getNode(node)

    def _getNode(self,node):
        if node in self.world:
            return self.world[node]
        else:
            iNewNode = self.nodeSeq
            self.nodeSeq += 1
            self.seq.append(node)
            self.world[node] = iNewNode
            return iNewNode

    def _add(self,inputs,output,nddr,at):
        if at == 0: # this is a leaf node
            if nddr<0:
                return output # return the output of the transition
            else:
                return nddr # return the node index
        if nddr in self.cache:
            return self.cache[nddr]
        # replace the node entry at each input with the index of the node from a recursive call to the next level down
        ### AKT: this code causes syntax error in Python 2.3:
        ### node = tuple( [at] + [ self._add(inputs,output,self.seq[nddr][i+1],at-1) if i in inputs[at-1] \
        ###                        else self.seq[nddr][i+1] for i in range(self.numStates) ] )
        temp = []
        for i in range(self.numStates):
            if i in inputs[at-1]:
                temp.append( self._add(inputs,output,self.seq[nddr][i+1],at-1) )
            else:
                temp.append( self.seq[nddr][i+1] )
        node = tuple( [at] + temp )
        r = self._getNode(node)
        self.cache[nddr] = r
        return r

    def _recreate(self,oseq,nddr,lev, swap_at):
        if lev == 0:
            return nddr
        if nddr in self.cache:
            return self.cache[nddr]
        if lev == swap_at:
            nodes = [lev]
            for j in range(self.numStates):
                node = tuple( [lev-1] + [ self._recreate(oseq, oseq[oseq[nddr][i+1]][j+1],lev-2,swap_at) for i in range(self.numStates) ] )
                r = self._getNode(node)
                nodes.append(r)
            node = tuple(nodes)
        else: 
            # each node entry is the node index retrieved from a recursive call to the next level down
            node = tuple( [lev] + [ self._recreate(oseq,oseq[nddr][i+1],lev-1,swap_at) for i in range(self.numStates) ] )
        r = self._getNode(node)
        self.cache[nddr] = r
        return r

    def _shrink(self, swap_at=-1):
        self.world = {}
        oseq = self.seq
        self.seq = []
        self.cache = {}
        self.nodeSeq = 0 ;
        self.curndd = self._recreate(oseq, self.curndd, self.numInputs, swap_at)
        self.shrinksize = len(self.seq) * 2

    def add_rule(self,inputs,output):
        self.cache = {}
        self.curndd = self._add(inputs,output,self.curndd,self.numInputs)
        if self.nodeSeq > self.shrinksize:
            self._shrink()

    def _setdefaults(self,nddr,off,at):
        if at == 0:
            if nddr<0:
                return off
            else:
                return nddr
        if nddr in self.cache:
            return self.cache[nddr]
        # each node entry is the node index retrieved from a recursive call to the next level down
        node = tuple( [at] + [ self._setdefaults(self.seq[nddr][i+1],i,at-1) for i in range(self.numStates) ] )
        node_index = self._getNode(node)
        self.cache[nddr] = node_index
        return node_index

    def _setDefaults(self):
        self.cache = {}
        self.curndd = self._setdefaults(self.curndd, -1, self.numInputs)

    def write(self, nhood):
        self._setDefaults()
        self._shrink()
        list_of_lines = ["@TREE\n"]
        list_of_lines.append("num_states=" + str(self.numStates) + '\n')
        list_of_lines.append("num_neighbors=" + repr(nhood) + '\n')
        list_of_lines.append("num_nodes="  + str(len(self.seq))  + '\n')
        for rule in self.seq:
            list_of_lines.append(' '.join(map(str,rule))+'\n')
        return list_of_lines

    def _treecompare(self, other, selfnode, othernode, memdict=None):

        if memdict is None:
            memdict = set([])

        if (selfnode, othernode) in memdict:
            return True

        sn = self.seq[selfnode]
        on = other.seq[othernode]

        lev = sn[0]

        if lev == 1:
            return (sn == on)

        for (sni, oni) in zip(sn[1:], on[1:]):
            if not self._treecompare(other, sni, oni, memdict):
                return False

        memdict.add((selfnode, othernode))
        return True

    def __eq__(self, other):
        '''
        Extensional comparison for equality; assumes trees are optimised.
        '''

        if self.numInputs != other.numInputs:
            return False

        if self.numStates != other.numStates:
            return False

        return self._treecompare(other, self.curndd, other.curndd)

    def invariant(self, perm):

        other = RuleTree(self.numStates, self.numInputs, initialise=False)
        other.seq = list(self.seq)
        other._shrink()

        perm = list(perm)
        for _ in range(len(perm)):
            for i in range(len(perm) - 1):
                if perm[i+1] < perm[i]:
                    perm[i], perm[i+1] = perm[i+1], perm[i]
                    other._shrink(swap_at=(len(perm) - i))

        return (other == self)

    def __ne__(self, other):

        return not self.__eq__(other)


def TransitionsToTree(n_states, nhood, transitions):
    '''Convert a set of transitions directly to a rule tree.'''

    numInputs = GetNumberOfInputs(nhood)
    tree_order = nhood[:numInputs][::-1] + nhood[numInputs:]
    tree = RuleTree(n_states, numInputs)
    for t in transitions:
        result = sum([(x[0] << (i << 3)) for (i, x) in enumerate(t[numInputs:])])
        tree.add_rule(t[:numInputs], result)

    return tree.write(tree_order)




import os
import re
from importlib import import_module
from .genuslist import genus_list
from .exceptions import NonCanonicalError, SurplusTreeError
from .rulefiles import rule2files

def obtain_genus(rulestring):

    for g in genus_list:
        m = re.match(g['regex'] + '$', rulestring)
        if m is not None:
            return g['name']

    raise ValueError('Rule "%s" does not belong to any genus' % rulestring)

def genus_to_module(genus):

    m = import_module('.' + genus, __name__)

    return m

def rule_property(rulestring, attribute):

    m = genus_to_module(obtain_genus(rulestring))
    attr = getattr(m, attribute)
    if callable(attr):
        attr = attr(rulestring)
    return attr

def create_rule(rulestring):

    rule_property(rulestring, 'create_rule')

def get_ltl_dict(rulestring):

    for i in range(10):
        d = str(i)
        rulestring = rulestring.replace(','+d,'+'+d)

    rulestring = rulestring.replace('-', '..')
    rparts = rulestring.lower().split(',')
    d = {'c': 2, 'm': 0, 'n': 'm'}

    for r in rparts:
        if (r[0] >= 'a') and (r[0] <= 'z'):
            d[r[0]] = r[1:]
        else:
            raise ValueError("Invalid LtL-like format")

    if d['n'] != 'm':
        raise ValueError("Unsupported LtL neighbourhood type")

    for k in 'rcm':
        d[k] = int(d[k])

    for k in 'bs':
        delta = (1 - d['m']) if (k == 's') else 0
        things = d[k].split('+')
        things = [(tuple(x.split('..')) if ('..' in x) else (x, x)) for x in things]
        d[k] = [(int(a) + delta, int(b) + delta) for (a, b) in things]

    return d

def decode_ltl(rulestring):

    if re.match("^[0-9,]*$", rulestring):
        # Kellie Evans format:
        return 'r%sb%st%ss%st%s' % tuple(rulestring.split(','))

    d = get_ltl_dict(rulestring)
    print("ltl_dict = %s" % str(d))

    if (len(d['b']) == 1) and (len(d['s']) == 1):
        rulestring = 'r%db%dt%ds%dt%d' % ((d['r'],) + d['b'][0] + d['s'][0])
    else:
        raise NotImplementedError("LifeViewer-style HROT notation is currently unsupported")

    if d['c'] > 2:
        rulestring = 'g%d%s' % (d['c'], rulestring)

    return rulestring

def sanirule(rulestring, drop_history=False):

    if ',' in rulestring:
        rulestring = decode_ltl(rulestring)

    if '.' in rulestring:
        try:
            rulestring = rule2files(rulestring)
            if (rulestring[0] != 'x'):
                raise ValueError("Rules specified by table/tree/code files must begin with a capital letter.")
        except SurplusTreeError as e:
            rulestring = str(e)

    rulestring = rulestring.lower()

    if (len(rulestring) > 7) and (rulestring[-7:] == 'history'):
        return sanirule(rulestring[:-7], drop_history) + ('' if drop_history else 'History')

    if '/' in rulestring:
        rparts = rulestring.split('/')
        if (len(rparts) <= 3):
            p = ['s', 'b', 'g'] # default parsing order == survivals/births/generations
            d = {}
            for r in rparts:
                if (len(r) > 0) and (r[0] in p):
                    d[r[0]] = r[1:]
                    p = [x for x in p if (x != r[0])]
                else:
                    d[p[0]] = r
                    p = p[1:]
            if ('g' in d) and (int(d['g']) >= 3):
                rulestring = 'g%sb%ss%s' % (d['g'], d['b'], d['s'])
            else:
                rulestring = 'b%ss%s' % (d['b'], d['s'])
        else:
            rulestring = rulestring.replace('/', '')

    commonnames = {'life': 'b3s23', 'pedestrianlife': 'b38s23', 'drylife': 'b37s23', 'highlife': 'b36s23'}

    if rulestring in commonnames:
        rulestring = commonnames[rulestring]

    return rulestring



'''
The following code was written by M. I. Wright (for Python 3) and
subsequently modified to be compatible with both versions of Python.
'''

from ._wright import *

def replace_bind(transition, pre, sub='', count=0):
    cop = []
    for i in map(int, transition):
        if not i:
            cop.append(0)
            continue
        cop.append('{}{}_{}'.format(pre, sub, count))
        count += 1
    return cop


def write_table(fp, rname, n_states, n_live, d_vars, transitions):
    fp.write('@RULE {}\n@TABLE\n'.format(rname))
    fp.write('n_states:{}\nneighborhood:Moore\nsymmetries:rotate4reflect\n'.format(n_states))
    # Variables
    live = range(1, n_states)
    for sub, (state, count) in d_vars.items():
        range_ = {i for i in live if i != state}
        if not range_:
            continue
        fp.write('\nvar not_{}_0 = {}'.format(sub, list(range_)).replace('[','{').replace(']','}'))
        for n in range(1, count):
            fp.write('\nvar not_{0}_{1} = not_{0}_0'.format(sub, n))
    fp.write('\nvar any_0 = {}'.format(list(range(n_states))).replace('[','{').replace(']','}'))
    for n in range(9):
        fp.write('\nvar any_{} = any_0'.format(n))
    fp.write('\nvar live_0 = {}'.format(list(live)).replace('[','{').replace(']','}'))
    for n in range(1, n_live):
        fp.write('\nvar live_{} = live_0'.format(n))
    fp.write('\n')
    # Transitions
    for tr in transitions:
        fp.write('\n' + ','.join(map(str, tr)))


def create_rule(rulestring):

    transitions = []
    bs = rulestring.split('d')[0]
    PERM = (rulestring[-1] == 'p')

    str2tab(bs) # Raise a NonCanonicalError if initial segment is non-canonical

    birth, survival = tuple(map(str.strip, bs[1:].split('s')))
    birth, survival = combine_rstring(birth), combine_rstring(survival)
    n_live, sum_len = 0, 2
    for counter, (num, subs) in enumerate(birth.items()):
        if num == '0':
            transitions.append([0]*9 + [1])
        elif num == '8':
            transitions.append([0] + ['live_{}'.format(i) for i in range(8)] + [1])
            n_live = 8
        else:
            transitions += [[0] + replace_bind(NAPKINS[num][sub], 'not_', num+sub) + [idx] for idx, sub in enumerate(subs, sum_len)]
            sum_len = 1 + transitions[-1][-1]

    d_vars = {k: (tr[-1], sum(1 for i in tr if isinstance(i, str))) for k, tr in zip((n+j for n, li in birth.items() for j in N_HOODS[:len(NAPKINS[n])] if not li or j in li), transitions)}
    transitions.append('')
    
    END = 'live_0' if PERM else 1  # Bind if rule is to be 'permanently deficient'
    for num, subs in survival.items():
        if num == '0':
            transitions.append(['live_0'] + [0]*8 + [END])
        elif num == '8':
            transitions.append(['live_{}'.format(i) for i in range(9)] + [END])
            n_live = 9
        else:
            transitions += [['live_0'] + replace_bind(NAPKINS[num][sub], 'live', count=1) + [END] for sub in subs]

    if survival:
        n_live = max(n_live, *(sum(1 for i in tr if isinstance(i, str)) for tr in transitions[1+transitions.index(''):]))

    n_states = 1 + max(t[-1] for t in transitions if t and isinstance(t[-1], int))
    transitions.append(['any_{}'.format(i) for i in range(9)] + [0])

    # Write temporary .rule file:
    with open((rulestring+'.rule'), 'w') as fp:
        write_table(fp, rulestring, n_states, n_live, d_vars, transitions)

    # Convert .rule file into C/C++ code:
    rule2files(rulestring + '.rule')

    # Link code into lifelib:
    make_header(rulestring)




import unittest

class TestCygwin(unittest.TestCase):

    def test_cygdir(self):

        import lifelib
        lifelib.add_cygdir('something')
        from lifelib import autocompile
        self.assertEqual(autocompile.cygwin_dirs, ['something'])

    def test_numpy(self):

        from lifelib import autocompile
        np_version = autocompile.verify_installation()

if __name__ == '__main__':
    unittest.main()



'''
Brian Prentice's BSFKL rulespace
'''

from .lifelike import create_rule as create_lifelike

family = 2
mantissa = {0, 1}
bitplanes = 2
number_of_states = 3


def create_rule(rulestring):

    parts = rulestring.replace('b', '').replace('s', ' ').replace('f', ' ').replace('k', ' ').replace('l', ' ').split(' ')

    bs_rule = 'b' + parts[0] + 's' + parts[1]
    f0_rule = 'b' + parts[2] + 's'
    k0_rule = 'b' + parts[3] + 's'
    l0_rule = 'b' + parts[4] + 's'

    create_lifelike(bs_rule)
    create_lifelike(f0_rule)
    create_lifelike(k0_rule)
    create_lifelike(l0_rule)

    with open('iterators_%s.h' % rulestring, 'w') as f:
        f.write('#pragma once\n')
        f.write('#include <stdint.h>\n')
        f.write('#include <cstdlib>\n')
        f.write('#include <iostream>\n')
        f.write('#include "../eors.h"\n')
        f.write('#include "../lifeperm.h"\n')

        for r in [bs_rule, f0_rule, k0_rule, l0_rule]:
            f.write('#include "iterators_%s.h"\n' % r)

        f.write('namespace %s {\n\n' % rulestring.replace('-', '_'))

        f.write('    bool iterate_var_leaf(uint64_t *inleaves, uint64_t *hleaves, uint64_t *outleaf) {\n\n')
        f.write('        %s::iterate_var_leaf(1, inleaves, outleaf);\n' % bs_rule)
        f.write('        outleaf[0] &= (~hleaves[3]); outleaf[1] &= (~hleaves[6]); outleaf[2] &= (~hleaves[9]); outleaf[3] &= (~hleaves[12]);\n')
        # At this point outleaf contains cells using B/S conditions.

        f.write('        outleaf[4] = inleaves[3] & (~outleaf[0]); outleaf[5] = inleaves[6]  & (~outleaf[1]);\n')
        f.write('        outleaf[6] = inleaves[9] & (~outleaf[2]); outleaf[7] = inleaves[12] & (~outleaf[3]);\n')
        # State-1 cells mature into state-2 cells.

        f.write('        uint64_t templeaf[4] = {0ull};\n')

        f.write('        %s::iterate_var_leaf(1, hleaves, templeaf);\n' % k0_rule)
        f.write('        templeaf[0] &= inleaves[3]; templeaf[1] &= inleaves[6]; templeaf[2] &= inleaves[9]; templeaf[3] &= inleaves[12];\n')
        f.write('        outleaf[0] &= (~templeaf[0]); outleaf[1] &= (~templeaf[1]); outleaf[2] &= (~templeaf[2]); outleaf[3] &= (~templeaf[3]);\n')
        f.write('        outleaf[4] &= (~templeaf[0]); outleaf[5] &= (~templeaf[1]); outleaf[6] &= (~templeaf[2]); outleaf[7] &= (~templeaf[3]);\n')
        # Effect of 'K' cells applied.

        f.write('        %s::iterate_var_leaf(1, hleaves, templeaf);\n' % f0_rule)
        f.write('        outleaf[0] &= (inleaves[3]  | templeaf[0]);\n')
        f.write('        outleaf[1] &= (inleaves[6]  | templeaf[1]);\n')
        f.write('        outleaf[2] &= (inleaves[9]  | templeaf[2]);\n')
        f.write('        outleaf[3] &= (inleaves[12] | templeaf[3]);\n')
        # We need 'F' for birth.

        f.write('        %s::iterate_var_leaf(1, inleaves, templeaf);\n' % l0_rule)
        f.write('        outleaf[4] |= (hleaves[3]  & (~templeaf[0]));\n')
        f.write('        outleaf[5] |= (hleaves[6]  & (~templeaf[1]));\n')
        f.write('        outleaf[6] |= (hleaves[9]  & (~templeaf[2]));\n')
        f.write('        outleaf[7] |= (hleaves[12] & (~templeaf[3]));\n')
        # 'L' causes state-2 cells to die; otherwise they remain state-2.

        f.write('        return false;\n')
        f.write('    }\n')
        f.write('}\n')



#!/usr/bin/python3

from sys import argv, stderr
import hashlib
import struct
from datetime import datetime as dt

from collections import OrderedDict

'''
We initially attempt to import sha3. On Python 2, this monkey-patches
the hashlib library to contain an implementation of sha3_256 (Keccak),
one of the cryptographic hashes used in the blockchain.
'''

try:
    import sha3
except ImportError as e:
    pass

sha2 = hashlib.sha256    # Merkle-Damgard
try:
    sha3 = hashlib.sha3_256  # Keccak
except AttributeError as e:
    stderr.write('\033[31;1mWarning:\033[0m SHA3 algorithm not present; ' +
    'please upgrade to Python 3.6 or pip install pysha3\n')
    sha3 = None

'''
We represent 256-bit binary blobs by padding them with a 32-bit CRC
checksum (to increase it to 288 bits) and represent them by 48 characters
of URL-safe base64.
'''

def crc32_for_byte(r):
    for j in range(8):
        r = (0 if (r & 1) else 0xEDB88320) ^ r >> 1;
    return r ^ 0xFF000000;

table = [crc32_for_byte(x) for x in range(256)]

def crc32(data):
    crc = 0
    for d in data:
        if not isinstance(d, int):
            d = ord(d)
        crc = table[(crc & 255) ^ d] ^ crc >> 8;
    return crc

def human_readable(data):
    '''
    This is a direct translation of the C++ code in coincludes/cryptography.h
    '''

    dig32 = list(struct.unpack('<IIIIIIII', data))
    dig32.append(crc32(data))
    digest = struct.pack('<IIIIIIIII', *dig32)
    digest = [(b if isinstance(b, int) else ord(b)) for b in digest]

    x = "";

    for i in range(0, 36, 3):
        a = 0;
        a += digest[i+2]; a <<= 8;
        a += digest[i+1]; a <<= 8;
        a += digest[i+0];
        for j in range(4):
            x += "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_-"[a & 63];
            a >>= 6;

    return x;

def block_rle(x):

    pn = struct.unpack('<Q', x[64:72])[0]
    seedroot = human_readable(x[32:64])
    prevseed = seedroot + str(pn)

    s  = '#C seed: %s\n' % prevseed
    s += '#C sha2: %s\n' % sha2(prevseed.encode('utf8')).hexdigest()

    digest = [(b if isinstance(b, int) else ord(b)) for b in sha2(prevseed.encode('utf8')).digest()]
    bits = [((b >> (7-i)) & 1) for b in digest for i in range(8)]
    rle = ''.join(['bo'[b] for b in bits])
    rle = tuple([rle[i:16+i] for i in range(0, 256, 16)])
    rle = '%s$%s$%s$%s$\n%s$%s$%s$%s$\n%s$%s$%s$%s$\n%s$%s$%s$%s!\n' % rle

    s += 'x = 16, y = 16, rule = B3/S23\n'
    s += rle

    return s

def block_info(x):

    bi = struct.unpack('<Q', x[80:88])[0]
    od = OrderedDict()
    od['idx'] = bi

    od['hex'] = OrderedDict()
    l = ['prev_sha2', 'prev_sha3', 'misc_info', 'addr_full',
         'txmr_sha2', 'txmr_sha3', 'xtra_nonc', 'addr_mine']

    for (i, r) in enumerate(l):

        try:
            # Python 3:
            h = x[32*i:32*(i+1)].hex()
        except AttributeError:
            # Python 2:
            h = ''.join("{:02x}".format(ord(c)) for c in x[32*i:32*(i+1)])

        od['hex'][r] = h

    pn = struct.unpack('<Q', x[64:72])[0]
    ts = struct.unpack('<Q', x[72:80])[0]
    ds = dt.utcfromtimestamp(ts//(10**9)).strftime('%Y-%m-%dT%H:%M:%S')
    ds += ('.' + str((10**9) + (ts % (10**9)))[1:])

    od['misc'] = OrderedDict()
    od['misc']['prev_nonce'] = pn
    od['misc']['time_stamp'] = ts
    od['misc']['block_time'] = ds
    od['misc']['blockindex'] = bi
    od['misc']['difficulty'] = struct.unpack('<d', x[88:96])[0]

    od['addresses'] = OrderedDict()
    od['addresses']['addr_full'] = human_readable(x[96:128])
    od['addresses']['addr_mine'] = human_readable(x[224:256])

    od['hashes'] = OrderedDict()
    od['hashes']['this_sha2'] = sha2(x).hexdigest()
    if sha3 is not None:
        od['hashes']['this_sha3'] = sha3(x).hexdigest()

    return od

def display_block(x):

    info = block_info(x)

    s = 'Block %d\n' % info['idx']
    s += ('=' * 64 + '\n\n')

    for (t, k) in [('Hexadecimal representation', 'hex'),
                   ('Miscellaneous information', 'misc'),
                   ('Winner addresses', 'addresses'),
                   ('Block hashes', 'hashes'),
                   ('Previous soup', 'nonkey')]:

        s += ('\n%s:\n' % t)
        s += (('-' * len(t)) + '-\n')
        if k in info:
            for item in info[k].items():
                s += '%s|%s\n' % item

    s += block_rle(x)

    s += ('\n' + ('=' * 64) + '\n')
    return s

def extract_blocks(filename):

    with open(filename, 'rb') as f:
        x = f.read()

    lr = (len(x) & 255)
    if lr:
        x += bytes([0] * (256 - lr))

    for i in range(0, len(x), 256):
        yield x[i:i+256]

def main(*filenames):

    for filename in filenames:
        for x in extract_blocks(filename):
            yield display_block(x)

if __name__ == '__main__':
    for db in main(*(argv[1:])):
        print(db)



from .lowlevel import *
from .pattern import Pattern
from .samples import download_samples as dsamples
from .samples import validate_symmetry
import os
import atexit

from ..genera import sanirule

from .identify import *

floats = {}


class Lifetree(object):

    def __init__(self, session, memory=1000, n_layers=None):

        if n_layers is None:
            n_layers = session.default_planes

        self.session = session
        lifelib = session.lifelib
        self.lifelib = lifelib
        self.ptr = lifelib('CreateLifetree', memory, n_layers)
        self.n_layers = n_layers

        session.lifetree_ptrs.add(self.ptr)

    def load_timeline(self, filename, n):

        import numpy as np
        frames = np.zeros((n,), dtype=np.uint64, order='C')
        frames = self.lifelib('LoadTimelineMC', self.ptr, filename, n, frames)

        patterns = []

        for f in frames:
            if f != 0:
                patterns.append(Pattern(self.session, (int(f),), self))

        return patterns

    def save_timeline(self, patterns, filename, header="", footer="", exponent=0, startgen=0):

        import numpy as np
        n = len(patterns) + 2
        frames = np.zeros((n,), dtype=np.uint64, order='C')
        for (i, p) in enumerate(patterns):
            frames[i] = p.ptr[0]
        frames[n-1] = startgen
        filename = os.path.abspath(filename)
        self.lifelib('SaveTimelineMC', self.ptr, filename, header, footer, frames, exponent)

    def unify(self, *patterns):

        output = self.pattern()

        for (i, pat) in enumerate(patterns):

            this_layer = self.pattern()
            this_layer += pat
            output += (this_layer << i)

        return output

    def deunify(self, pattern):

        n = self.n_layers
        return tuple([((pattern << (n - 1 - i)) >> (n - 1)) for i in range(n)])

    def loadu(self, filename):
        '''
        Load an uncompressed pattern file.
        '''

        filename = os.path.abspath(filename)
        metadata = identify_metadata(filename)
        timeline = False

        if 'frames' in metadata:
            try:
                n = int(metadata['frames'][0])
                timeline = True
            except:
                pass

        if timeline:
            # timeline file
            return self.load_timeline(filename, n)
        else:
            # pattern file
            pptr = self.lifelib('CreatePatternFromFile', self.ptr, filename)
            pat = Pattern(self.session, pptr, self)
            pat.comments = metadata['comments']
            return pat

    def load(self, filename, compressed='deduce', tempfile='tempfile'):
        '''
        Load a (possibly compressed) pattern file.
        '''

        filename = os.path.abspath(filename)
        tempfile = os.path.abspath(tempfile)

        if not os.path.isfile(filename):
            raise OSError("%s does not exist or is not a regular file" % filename)

        if compressed == 'deduce':
            compressed = identify_compression(filename)

        if compressed:

            from importlib import import_module
            decompressor = import_module(compressed)

            import shutil
            with decompressor.open(filename, 'rb') as f_in:
                with open(tempfile, 'wb') as f_out:
                    shutil.copyfileobj(f_in, f_out)
            rt = self.loadu(tempfile)
            try:
                os.remove(tempfile)
            except OSError:
                pass
        else:
            rt = self.loadu(filename)

        return rt

    def hashsoup(self, rule, symmetry, seed):

        validate_symmetry(rule, symmetry)

        pptr = self.lifelib('Hashsoup', self.ptr, rule, symmetry, seed)
        return Pattern(self.session, pptr, self)

    def download_samples(self, apgcode, rule, **kwargs):

        return dsamples(self, apgcode, rule, **kwargs)

    def pattern(self, rle="", rule=None, tempfile='tempfile', verify_apgcode=True):

        tempfile = os.path.abspath(tempfile)
        is_apgcode = False

        if ('=' in rle) or ('[' in rle):
            # Headerful RLE/MC; save and reload:
            with open(tempfile, 'w') as f:
                f.write(rle)
            pptr = self.lifelib('CreatePatternFromFile', self.ptr, tempfile)
            try:
                os.remove(tempfile)
            except OSError:
                pass
        else:
            # Headerless RLE:
            if rule is None:
                if len(self.session.rules) == 1:
                    rule = self.session.rules[0]
                else:
                    raise TypeError("For headerless RLE, rule must be specified unless session has a unique rule")

            rule = self.session.verify_rule(rule)
            pptr = self.lifelib('CreatePatternFromRLE', self.ptr, rle, rule)

            is_apgcode = ('_' in rle) and ('!' not in rle)

        if is_apgcode and (rle[0] != 'x'):
            raise ValueError("Only apgcodes beginning with 'x' contain sufficient information to reconstruct the pattern.")

        pat = Pattern(self.session, pptr, self)

        if is_apgcode and verify_apgcode and not rle.startswith('xg'):
            prefix = rle.split('_')[0]
            prefint = int(prefix[2:])
            period = 1 if (prefix[1] == 's') else prefint

            if (pat[period].centre() != pat.centre()) or (pat.period != period):
                raise ValueError("Pattern does not have the period implied by its apgcode.")

            if (prefix[1] == 's') and (pat.population != prefint):
                raise ValueError("Pattern does not have the population implied by its apgcode.")

            if (verify_apgcode == 'canonical') and (rle != pat.apgcode):
                raise ValueError("The canonical form of %s is %s" % (rle, pat.apgcode))

        return pat

    def __del__(self):

        if self.ptr in self.session.lifetree_ptrs:
            self.lifelib('DeleteLifetree', self.ptr)
            self.session.lifetree_ptrs.remove(self.ptr)

all_sessions = []

def cleanup_sessions():

    # Deallocate patterns
    for sess in all_sessions:
        for ptr in sess.pattern_ptrs:
            sess.lifelib('DeletePattern', ptr)
        sess.pattern_ptrs = set([])

    # Deallocate lifetrees
    for sess in all_sessions:
        for ptr in sess.lifetree_ptrs:
            sess.lifelib('DeleteLifetree', ptr)
        sess.lifetree_ptrs = set([])

    # Wait for child processes
    for sess in all_sessions:
        sess.lifelib.annihilate()

    # Remove references
    while all_sessions:
        all_sessions.pop()

atexit.register(cleanup_sessions)

class Session(object):

    def newfloat(self, name):

        if name not in floats:
            floats[name] = 0

        floats[name] += 1
        return ('%s%d' % (name, floats[name]))

    def __init__(self, soname, rules=['b3s23'], local_bash=None,
                 local_python='python', default_planes=4):

        self.default_planes = default_planes
        self.rules = list(rules)
        self.lifelib = WrappedLibrary(soname, local_bash=local_bash, local_python=local_python)
        self.lifetree_ptrs = set([])
        self.pattern_ptrs = set([])

        all_sessions.append(self)

    def lifetree(self, *args, **kwargs):

        return Lifetree(self, *args, **kwargs)

    def verify_rule(self, rule):

        rule = sanirule(rule)

        if (len(rule) > 7) and (rule[-6:] == 'istory'):
            newrule = rule[:-7]
        else:
            newrule = rule

        if newrule not in self.rules:
            raise ValueError("Rule %s is not in the configured rules %s for this session" % (rule, self.rules))

        return rule



'''
Downloads CAViewer if not already downloaded.
'''

import os
import zipfile

DOWNLOAD_LINK = "https://gist.github.com/jedlimlx/a4ad9e4bddf1bc0fcfe220c4c150ff11/raw/02a656bc9dda3d6816d0bbcacb0c695b4032dc2c/CAViewer-Linux.zip"


def download():

    this_dir = os.path.dirname(os.path.abspath(__file__))
    zip_location = os.path.join(this_dir, 'CAViewer-Linux.zip')
    caviewer_path = this_dir + "/bin/CAViewer"

    try:
        from urllib import urlretrieve
    except ImportError:
        from urllib.request import urlretrieve

    if not os.path.exists(caviewer_path):  # Checking if CAViewer exists

        print("Downloading CAViewer...")
        urlretrieve(DOWNLOAD_LINK, zip_location)

        print("Unzipping...")
        with zipfile.ZipFile(zip_location, "r") as z:
            z.extractall(this_dir)

        print("Download complete!")

        os.chmod(caviewer_path, 0o755)




from .ltl import mantissa, family, bitplanes

def get_hrot_params(rulestring):

    rulerange = int(rulestring[1])
    dhexchars = rulerange * (rulerange + 1)
    gparams = rulestring[3:].replace('z', '').split('s')

    if len(gparams[0]) != dhexchars:
        raise ValueError("Birth conditions for range-%d rules must have exactly %d hex characters, not %d" % (rulerange, dhexchars, len(gparams[0])))

    if len(gparams[1]) != dhexchars:
        raise ValueError("Survival conditions for range-%d rules must have exactly %d hex characters, not %d" % (rulerange, dhexchars, len(gparams[1])))

    gparams = tuple([rulerange] + [int(x, 16) for x in gparams] + ['z' in rulestring])

    return gparams

def write_rule_data(f, b_int, s_int, s_zero):

    b_int = 2 * b_int
    s_int = 4 * s_int + (2 if s_zero else 0)

    births    = [(b_int >> (8 * i)) & 255 for i in range(16)]
    survivals = [(s_int >> (8 * i)) & 255 for i in range(16)]

    marray = [b ^ s for (b, s) in zip(births, survivals)]
    marray += [255 ^ s for s in survivals]

    marray += ([15] * 16)
    marray += [1, 2, 4, 8, 16, 32, 64, 128]
    marray += [1, 2, 4, 8, 16, 32, 64, 128]

    f.write('const static uint8_t ruledata[] __attribute__((aligned(64))) = {%d,\n' % marray[0])
    for i in range(3):
        currstring = '   '
        for j in range(i*21+1, i*21+22):
            currstring += ((' %3d};' if (j == 63) else ' %3d,') % marray[j])
        f.write(currstring + '\n')


def create_rule(rulestring):

    rulerange, b_int, s_int, s_zero = get_hrot_params(rulestring)

    with open('iterators_%s.h' % rulestring, 'w') as f:
        f.write('#pragma once\n')
        f.write('#include <stdint.h>\n')
        f.write('#include "../ltl.h"\n')
        f.write('namespace %s {\n\n' % rulestring.replace('-', '_'))

        write_rule_data(f, b_int, s_int, s_zero)

        f.write('    bool iterate_var_leaf(uint64_t *inleaves, uint64_t *outleaf) {\n\n')
        f.write('        apg::hrot_kernel(inleaves, outleaf, %d, ruledata);\n' % rulerange)
        f.write('        return false;\n')
        f.write('    }\n')
        f.write('}\n')



import unittest
import lifelib
import os

oca_rules = ['g4b2ikn3ais345', 'b36s23', 'g3b3s23', 'r7b65t95s65t114', 'g11b3s23', 'r2b1e7108s3e7c80', 'b347s3456f038k234567l0346']

class TestOtherRules(unittest.TestCase):

    @classmethod
    def setUpClass(cls):

        lifelib.load_rules(*oca_rules, force_compile=True)

    def setUp(self):

        self.sess = lifelib.load_rules(*oca_rules)

    def test_bsfkl(self):

        lt = self.sess.lifetree()
        apgcode = 'xq20_xeefz77_y2g5zw55'
        ss = lt.pattern(apgcode, 'b347s3456f038k234567l0346')
        self.assertEqual(ss.period, 20)

    def test_uint64(self):

        try:
            os.makedirs('lifelib/tempdir')
        except:
            pass

        try:

            lt = self.sess.lifetree(n_layers=64)
            apgcode = 'xp32_ss2gzw11_wsfm76cg_x86668_wke2024_ws6i1_x1'
            rro = lt.pattern(apgcode, 'b36s23', verify_apgcode=False)
            rro[0, 0] = 10 ** 19
            self.assertEqual(rro[0, 0], 10 ** 19)

            # Test multistate RLE
            rro.save('lifelib/tempdir/rrom.rle')
            rro2 = lt.load('lifelib/tempdir/rrom.rle')
            self.assertEqual(rro2[0, 0], 10 ** 19)
            self.assertEqual(rro, rro2)

            # Test multistate MC
            rro.save('lifelib/tempdir/rrom.mc')
            rro3 = lt.load('lifelib/tempdir/rrom.mc')
            self.assertEqual(rro3[0, 0], 10 ** 19)
            self.assertEqual(rro, rro3)

            with open('lifelib/tempdir/rrom.rle') as f:
                for a in f:
                    a = a.strip()
                    if a:
                        lastline = a

            self.assertEqual(lastline, "xzwqrxqvyxuwpuwzwP2.pJ.R$2.ApBpFFN$2A2pB3FJ$2ARN3.F$2ApBAR3.B$2.2A!")

        finally:
            import shutil
            # shutil.rmtree('lifelib/tempdir')

    def test_multistate_apgcode(self):

        lt = self.sess.lifetree(memory=1000, n_layers=8)
        apgcode = 'xp32_ss2gzw11_wsfm76cg_x86668_wke2024_ws6i1_x1'
        rro = lt.pattern(apgcode, 'g11b3s23')
        self.assertEqual(rro.apgcode, apgcode)

        rrorle = lt.pattern('''x = 8, y = 7, rule = 23/3/11
4.B$2.2AEC$.2A3ED$ABCHJHE$.2AB.2HF$2.AJ2.G$7.J!''', 'g11b3s23')

        self.assertEqual(rrorle.apgcode, apgcode)

    def test_highlife(self):

        lt = self.sess.lifetree(memory=1000, n_layers=1)
        rep = lt.pattern('3o$o2bo$o3bo$bo2bo$2b3o!', 'b36s23')
        self.assertEqual(rep[12 * (2 ** 40 - 1)].population, 12 * (2 ** 40))
        self.assertEqual(rep[12 * (2 ** 40)].population, 24)

    def test_hrot(self):

        lt = self.sess.lifetree(memory=1000, n_layers=2)
        p61 = lt.pattern('2o$2o9$20b2o$20b2o!', 'r2b1e7108s3e7c80')
        self.assertEqual(p61.apgcode, 'xp61_33zzyg33')

    def test_generations(self):

        lt = self.sess.lifetree(memory=1000, n_layers=2)
        p54soup = lt.pattern('''oobbooobooobooob$
bobbboooobooobbo$
bbbbobobbooobobo$
oboboboooobooboo$
obbbobbboobooobo$
bobobbboobbboobb$
ooooooobobooboob$
bboooboooobobboo$
boooobobbooobbbb$
bbbobbobobboboob$
oooboboobooboooo$
boobobbooobooobb$
obobbobbobbbbbbo$
oobbbbbbbbbooobo$
bbbboboobbooboob$
bbbbooobobobbboo!''', 'g3b3s23')
        self.assertEqual(p54soup.oscar(verbose=False)['period'], 54)

        x = p54soup.population
        self.assertEqual(p54soup[30, 7], 0)
        p54soup[30, 7] = 2
        self.assertEqual(p54soup[30, 7], 2)
        self.assertEqual(p54soup.population, x + 1)

        try:
            import numpy as np
        except ImportError:
            return

        coords = np.array([[0, 1], [2, 3], [4, 5], [-6, 69], [73, -48]], dtype=np.int64)
        values = np.array([1, 0, 2, 1, 2], dtype=np.int64)

        p54soup[coords] = values
        values2 = p54soup[coords]

        self.assertEqual(np.all(values == values2), True)

    def test_isogeny(self):

        lt = self.sess.lifetree(memory=1000, n_layers=4)
        p552soup = lt.pattern('''4.A.2A.A.2A$2A3.3A3.4A$2.2A2.A.A.6A$A2.5A2.6A$.2A.2A2.A4.A.A$2A.A2.2A
.A.3A$.A.2A.3A.2A.3A$2A.2A2.A.A.A.3A$A.2A.3A4.2A.A$A3.A3.6A$2A4.A3.A
4.A$.3A5.6A$5.A.A.A2.A2.A$.A3.3A.A.2A.2A$A.A.A.2A.5A.A$3A2.A2.A2.A.A.
A!''', 'g4b2ikn3ais345')
        self.assertEqual(p552soup.oscar(verbose=False)['period'], 552)

    def test_largerthanlife(self):

        lt = self.sess.lifetree(memory=1000, n_layers=1)
        rrosoup = lt.pattern('''booobbbboobboooo$
bbooobooobobbobo$
bboobboobooobboo$
oobobobobbbooboo$
booooobboboooobo$
booboboboobooooo$
obboooobbbbbobbo$
oooooooobobboobb$
obbbbbobbbbboobb$
bobobboooooobobb$
bbbbobbbbooobboo$
bobobbbobbobboob$
obbobbbboobbobbb$
bobobbooooooooob$
oobobboooboobboo$
bboboobbobbbbobb!''', 'r7b65t95s65t114')
        self.assertEqual(rrosoup.oscar(verbose=False)['period'], 552)

if __name__ == '__main__':
    unittest.main()



family = 4
bitplanes = 1
mantissa = {0, 1}

def get_ltl_params(rulestring):

    gparams = rulestring[1:].replace('b', 't').replace('s', 't').split('t')
    gparams = tuple(map(int, gparams))
    if (gparams[1] > gparams[2]):
        raise ValueError("Minimum birth count cannot exceed maximum birth count")
    if (gparams[3] > gparams[4]):
        raise ValueError("Minimum survival count cannot exceed maximum survival count")
    if (gparams[2] >= ((2 * gparams[0] + 1) ** 2)):
        raise ValueError("Maximum birth count cannot exceed or equal neighbourhood area")
    if (gparams[4] > ((2 * gparams[0] + 1) ** 2)):
        raise ValueError("Maximum survival count cannot exceed neighbourhood area")

    return gparams

def create_rule(rulestring):

    gparams = get_ltl_params(rulestring)

    with open('iterators_%s.h' % rulestring, 'w') as f:
        f.write('#pragma once\n')
        f.write('#include <stdint.h>\n')
        f.write('#include "../ltl.h"\n')
        f.write('namespace %s {\n\n' % rulestring.replace('-', '_'))

        f.write('    bool iterate_var_leaf(uint64_t *inleaves, uint64_t *outleaf) {\n\n')
        f.write('        apg::ltl_kernel(inleaves, outleaf')
        for g in gparams:
            f.write(', %d' % g)
        f.write(');\n')
        f.write('        return false;\n')
        f.write('    }\n')
        f.write('}\n')
